module mod_outer


use mod_spline
use mod_policies
use mod_inner
implicit none


contains


subroutine outer(Vmat,kappakp,kappaw,kappar,agrid,mgrid,kgrid,invTa,invTm,zgrid,pzij,bgrid,xgrid,pbij,pxij,idmat,izvec,simTT,Yss, &
	aggsim,disaggsim,mumat,zvec,xghz,wghz, &
	ALPHA,DELTA,HBAR,ALOW,B0,BETA,RHOZ,kappa1,kappa2,Trscale,Transprog1,Transprog2,T0, &
	tol,tolmkt,outermkt,foreind,bsctr,bspct,maxcountmkt,calcma,bsfixr,diagnum,fcsteqn,linflag,hsvflag,conzflag)

	integer, parameter :: calcdist = 1

	real(8), intent(in) :: Vmat(:,:,:,:), kappakp(:,:), kappaw(:,:), kappar(:,:), agrid(:), mgrid(:), kgrid(:), &
	invTa(:,:), invTm(:,:), zgrid(:), pzij(:,:), bgrid(:), xgrid(:), pbij(:,:), pxij(:,:), Yss, zvec(:), xghz(:), wghz(:)
	integer, intent(in) :: idmat(:,:), izvec(:), simTT
	real(8), intent(out) :: aggsim(:,:), disaggsim(:,:,:)
	real(8), intent(inout) :: mumat(:,:)

	real(8), intent(in) :: ALPHA, DELTA, HBAR, ALOW, B0, BETA, RHOZ, kappa1, kappa2, Trscale, Transprog1, Transprog2, T0
    real(8), intent(in) :: tol, tolmkt, bspct
	integer, intent(in) :: outermkt, foreind, bsctr, maxcountmkt, calcma, bsfixr, diagnum, fcsteqn, linflag, hsvflag, conzflag

	real(8), allocatable :: Tmumat(:,:), hdist(:,:), cdist(:,:), ldist(:,:), taxdist(:,:), transdist(:,:), minasseta(:), minassetd(:)
    real(8), allocatable :: cfmat0(:,:,:,:,:), EVmat(:,:), EVmat0(:,:,:,:), EVmatz(:,:,:), EVmate(:,:)
	real(8), allocatable :: mukx(:), consx(:), hx(:), taxx(:), transx(:), kx(:), lx(:)
	real(8) mukw(5), consw(5), hw(5), taxw(5), transw(5), kw(5), lw(5)
    integer na, nk, nm, ra, rm, nb, nx, ne, nz, ib, jb, ix, jx, ie, je, iz, jz, ik, time, countmkt, multhres, jghz, nghz
	real(8) znow, mnow, mp, wf, rf, w0, r0, Lagg, Ls, Cagg, Hs, Taxagg, Transagg, Ks, Ys, Xagg, Gagg, wfvec(simTT), w0vec(simTT), zp, wz
	real(8) damp, diffma, Kagg, KLs, KLagg
	integer ct1, ct2, cr, cto1, cto2, cro
    real(8) eptime


	na = size(agrid,1)
	nk = size(kgrid,1)
	nm = size(mgrid,1)
	nb = size(bgrid,1)
	nx = size(xgrid,1)
	ne = nb*nx
	nz = size(zgrid,1)
	nghz = size(xghz,1)
	allocate(Tmumat(nk,ne),hdist(nk,ne),cdist(nk,ne),ldist(nk,ne),taxdist(nk,ne),transdist(nk,ne),minasseta(nx),minassetd(nx))
	allocate(EVmatz(na,nm,ne),EVmate(na,nm))
	ra = na-2
	rm = nm-2
	allocate(mukx(nx),consx(nx),hx(nx),taxx(nx),transx(nx),kx(nx),lx(nx))


	print *, "  OUTER LOOP"
	call system_clock(cto1, cro)

	! compute conditional expected value functions on grid points
	if (conzflag==2) then

		! NOTE: conditional expectation is calculated in every time (z is continuous)
		allocate(EVmat(na,nm),EVmat0(na,nm,ne,1),cfmat0(16,ra+1,rm+1,ne,1))

	else

		allocate(EVmat(na,nm),EVmat0(na,nm,ne,nz),cfmat0(16,ra+1,rm+1,ne,nz))

	    do ie = 1,ne

			ib = idmat(ie,1)
			ix = idmat(ie,2)

	        do iz = 1,nz

				EVmat = 0.0d0
	            do jz = 1,nz

	                do je = 1,ne

						jb = idmat(je,1)
						jx = idmat(je,2)
	                    EVmat = EVmat + pbij(ib,jb)*pxij(ix,jx)*pzij(iz,jz)*reshape(Vmat(:,:,je,jz),(/na,nm/))

	                end do

				end do

				EVmat0(:,:,ie,iz) = EVmat
				if (linflag==0) cfmat0(:,:,:,ie,iz) = spfit2(invTa,invTm,EVmat,ra,rm,agrid,mgrid)

	        end do

	    end do

	end if


    do time = 1,simTT

		call system_clock(ct1, cr)
        mumat = mumat/sum(sum(mumat,1),1)

		if (conzflag==2) then

			! NOTE: conditional expectation is calculated in every time (z is continuous)
			znow = zvec(time)
            iz = 1

			! compute conditional expected value functions on grid points
		    do ie = 1,ne

				ib = idmat(ie,1)
				ix = idmat(ie,2)

				! nz is the number of gaussian quadrature
				EVmat = 0.0d0
				do jghz = 1,nghz

					zp = exp(RHOZ*log(znow) + xghz(jghz))
					call findjw(zp,zgrid,jz,wz)
					EVmatz = wz*reshape(Vmat(:,:,:,jz),(/na,nm,ne/)) + (1.0d0-wz)*reshape(Vmat(:,:,:,jz+1),(/na,nm,ne/))

					EVmate = 0.0d0
					do je = 1,ne

						jb = idmat(je,1)
						jx = idmat(je,2)
						EVmate = EVmate + pbij(ib,jb)*pxij(ix,jx)*reshape(EVmatz(:,:,je),(/na,nm/))

					end do

					EVmat = EVmat + wghz(jghz)*EVmate

				end do

				EVmat0(:,:,ie,1) = EVmat
				if (linflag==0) cfmat0(:,:,:,ie,1) = spfit2(invTa,invTm,EVmat,ra,rm,agrid,mgrid)

		    end do

		else

			iz = izvec(time)
	        znow = zgrid(iz)

		end if

        mnow = 0.0d0
        do ie = 1,ne
            mnow = mnow + sum(mumat(:,ie)*kgrid)
        end do

        if (conzflag==1) then

			mp = exp(kappakp(1,iz) + kappakp(2,iz)*log(mnow))
			if (fcsteqn==1) then
				wf = exp(kappaw(1,iz) + kappaw(2,iz)*log(mnow))
				rf = (znow**(1.0d0/ALPHA))*ALPHA*((wf/(1.0d0-ALPHA))**((ALPHA-1.0d0)/ALPHA)) - DELTA
			elseif (fcsteqn==2) then
				rf = exp(kappar(1,iz) + kappar(2,iz)*log(mnow)) - DELTA
				wf = (znow**(1.0d0/(1.0d0-ALPHA)))*(1.0d0-ALPHA)*(((rf+DELTA)/ALPHA)**(ALPHA/(ALPHA-1.0d0)))
			else
				wf = exp(kappaw(1,iz) + kappaw(2,iz)*log(mnow))
				rf = exp(kappar(1,iz) + kappar(2,iz)*log(mnow)) - DELTA
			end if

        elseif (conzflag==3) then

			mp = exp(kappakp(1,1) + kappakp(2,1)*log(mnow) + kappakp(3,1)*log(znow) + kappakp(4,1)*log(znow)*log(mnow))
			if (fcsteqn==1) then
				wf = exp(kappaw(1,1) + kappaw(2,1)*log(mnow) + kappaw(3,1)*log(znow) + kappaw(4,1)*log(znow)*log(mnow))
				rf = (znow**(1.0d0/ALPHA))*ALPHA*((wf/(1.0d0-ALPHA))**((ALPHA-1.0d0)/ALPHA)) - DELTA
			elseif (fcsteqn==2) then
				rf = exp(kappar(1,1) + kappar(2,1)*log(mnow) + kappar(3,1)*log(znow) + kappar(4,1)*log(znow)*log(mnow)) - DELTA
				wf = (znow**(1.0d0/(1.0d0-ALPHA)))*(1.0d0-ALPHA)*(((rf+DELTA)/ALPHA)**(ALPHA/(ALPHA-1.0d0)))
			else
				wf = exp(kappaw(1,1) + kappaw(2,1)*log(mnow) + kappaw(3,1)*log(znow) + kappaw(4,1)*log(znow)*log(mnow))
				rf = exp(kappar(1,1) + kappar(2,1)*log(mnow) + kappar(3,1)*log(znow) + kappar(4,1)*log(znow)*log(mnow)) - DELTA
			end if

        else ! conzflag = 0 or 2

			mp = exp(kappakp(1,1) + kappakp(2,1)*log(mnow) + kappakp(3,1)*log(znow))
			if (fcsteqn==1) then
				wf = exp(kappaw(1,1) + kappaw(2,1)*log(mnow) + kappaw(3,1)*log(znow))
				rf = (znow**(1.0d0/ALPHA))*ALPHA*((wf/(1.0d0-ALPHA))**((ALPHA-1.0d0)/ALPHA)) - DELTA
			elseif (fcsteqn==2) then
				rf = exp(kappar(1,1) + kappar(2,1)*log(mnow) + kappar(3,1)*log(znow)) - DELTA
				wf = (znow**(1.0d0/(1.0d0-ALPHA)))*(1.0d0-ALPHA)*(((rf+DELTA)/ALPHA)**(ALPHA/(ALPHA-1.0d0)))
			else
				wf = exp(kappaw(1,1) + kappaw(2,1)*log(mnow) + kappaw(3,1)*log(znow))
				rf = exp(kappar(1,1) + kappar(2,1)*log(mnow) + kappar(3,1)*log(znow)) - DELTA
			end if

        end if

		call bisect(agrid,kgrid,bgrid,xgrid,mgrid,znow,mnow,invTa,cfmat0,EVmat0,mumat,idmat,iz,mp, &
			ALPHA,DELTA,ALOW,B0,HBAR,kappa1,kappa2,Yss,Trscale,Transprog1,Transprog2,T0,pbij,pxij,tol,tolmkt,outermkt,foreind,bsctr,bspct,maxcountmkt,calcma,bsfixr,linflag,hsvflag,time, &
			wf,rf,w0,r0,Lagg,Kagg,Cagg,Hs,Taxagg,Transagg,Tmumat,cdist,hdist,taxdist,transdist,ldist,countmkt,multhres)

		wfvec(time) = wf
		w0vec(time) = w0

		Ks = mnow
		Ys = znow*(Ks**ALPHA)*(Lagg**(1.0d0-ALPHA))
        
		Gagg = Taxagg-Transagg
		Xagg = Ys-Cagg-Gagg

		if (calcdist) call calcdisaggstat(HBAR,idmat,kgrid,mumat,cdist,hdist,taxdist,transdist,ldist, &
			mukx,consx,hx,taxx,transx,kx,lx,mukw,consw,hw,taxw,transw,kw,lw)

		
		if (mod(time,diagnum)==0 .or. multhres==1) then

			if (multhres==1) then
				print *, '  NOTE: multiple thresholds are found' 
				if (calcma==1) print *, '  we are using the ones found by descending order search'
				if (calcma==2) print *, '  we are using the ones found by ascending order search'
			end if

	        write(*,"('  time      ', I5, '  ( z, m, w, r) = (', F8.5, ', ', F8.5, ', ', F8.5, ', ', F8.5, ')')") &
			time, znow, mnow, w0, r0
			write(*,"('  errors(%)        w : ', F10.5, ', r : ', F10.5, ', inv : ', F10.5)") &
	        log(w0/wf)*100.0d0, log(r0/rf)*100.0d0, log(Xagg/(mp-(1.0d0-DELTA)*mnow))*100.0d0
			write(*,"('  ( w, wf, r, rf) = (', F8.5, ', ', F8.5, ', ', F8.5, ', ', F8.5, ') ')") &
			w0, wf, r0, rf
            KLs = Ks/Lagg
            KLagg = (w0/(znow*(1.0d0-ALPHA)))**(1.0d0/alpha)
            write(*,"('  bisection ', I4, '  error(%) K/L = ', F20.15)") countmkt, abs(log(KLs)-log(KLagg))*100.0d0

			call system_clock(ct2, cr)
		    eptime = dble(ct2-ct1)/cr
		    write(*,"('  Elapsed time = ', F10.5)") eptime

		end if

		aggsim(time,1) = Ks
		aggsim(time,2) = w0
		aggsim(time,3) = r0
		aggsim(time,4) = Ys
		aggsim(time,5) = Cagg
		aggsim(time,6) = Xagg
		aggsim(time,7) = Lagg
		aggsim(time,8) = Hs
		aggsim(time,9) = znow
        aggsim(time,10) = Transagg

		disaggsim(time,:,1) = consx/mukx
        disaggsim(time,:,2) = hx/mukx
        disaggsim(time,:,3) = taxx/mukx
        disaggsim(time,:,4) = transx/mukx
        disaggsim(time,:,5) = kx/mukx
        disaggsim(time,:,6) = lx/mukx

        mumat = Tmumat

	end do ! time

    call system_clock(cto2, cro)
    eptime = dble(cto2-cto1)/cro
    write(*,"('  Outerloop elapsed time (minutes) = ', F10.5)") eptime/60


end subroutine outer


subroutine bisect(agrid,kgrid,bgrid,xgrid,mgrid,znow,mnow,invTa,cfmat0,EVmat0,mumat,idmat,iz,mp, &
		ALPHA,DELTA,ALOW,B0,HBAR,kappa1,kappa2,Yss,Trscale,Transprog1,Transprog2,T0,pbij,pxij,tol,tolmkt,outermkt,foreind,bsctr,bspct,maxcountmkt,calcma,bsfixr,linflag,hsvflag,time, &
		wf,rf,w0,r0,Lagg,Kagg,Cagg,Hs,Taxagg,Transagg,Tmumat,cdist,hdist,taxdist,transdist,ldist,countmkt,multhres)

	real(8), intent(in) :: agrid(:), kgrid(:), bgrid(:), xgrid(:), mgrid(:), invTa(:,:), cfmat0(:,:,:,:,:), EVmat0(:,:,:,:), mumat(:,:), znow, mnow, mp
	integer, intent(in) :: idmat(:,:), iz, outermkt, foreind, bsctr, time, maxcountmkt, calcma, bsfixr, linflag, hsvflag
	real(8), intent(in) :: ALPHA, DELTA, ALOW, B0, HBAR, kappa1, kappa2, Yss, Trscale, Transprog1,Transprog2, T0, pbij(:,:), pxij(:,:), tol, tolmkt, bspct
	real(8), intent(in) :: wf, rf
	real(8), intent(out) :: w0, r0, Lagg, Kagg, Cagg, Hs, Taxagg, Transagg, Tmumat(:,:), cdist(:,:), hdist(:,:), taxdist(:,:), transdist(:,:), ldist(:,:)
	integer, intent(out) :: countmkt, multhres
	real(8), parameter :: scale = 10.0d0
	real(8) distancemkt, rL, rH, wL, wH, LaggL, LaggH, g0, LsL, LsH, Ls, KaggL, KaggH, Ks, KLs, KLsL, KLsH, KLagg, KLaggL, KLaggH


	if (time==1) then
		if (calcma==1) print *, "  NOTE: minasset (with descending order search) is used to calculate distribution"
		if (calcma==2) print *, "  NOTE: minasset (with ascending order search) is used to calculate distribution"
		if (outermkt==1) then
			print *, "  NOTE: market-clearing is imposed"
			if (bsctr==1) then
				print *, "  NOTE: bisection is done for interest rate r"
				if (bsfixr==1) print *, "  NOTE: w is fixed at wf while bisecting r"
			else
				print *, "  NOTE: bisection is done for wage w"
				if (bsfixr==1) print *, "  NOTE: r is fixed at rf while bisecting w"
			end if
		else
			print *, "  NOTE: market-clearing is NOT imposed"
		end if
	end if

	distancemkt = 1d+4
	countmkt = 0
    
    Ks = mnow

	if ((outermkt==0) .or. (foreind==0)) then

		! no market clearing
		countmkt = countmkt + 1
		call pricemap(agrid,kgrid,bgrid,xgrid,mgrid,invTa,cfmat0,EVmat0,mumat,idmat,iz,mp,wf,rf,ALOW,B0,HBAR,kappa1,kappa2,Yss,Trscale,Transprog1,Transprog2,T0, &
			pbij,pxij,tol,countmkt,calcma,linflag,hsvflag,Cagg,Hs,Taxagg,Transagg,Ls,Tmumat,cdist,hdist,taxdist,transdist,ldist,multhres)
		
		Lagg = Ls
		w0 = znow*(1.0d0-ALPHA)*(mnow/Lagg)**ALPHA
		r0 = znow*ALPHA*(mnow/Lagg)**(ALPHA-1.0d0) - DELTA

	else

		! market clearing by bisection method (using interest rate)
		if (bsctr==1) then
			! NOTE: Initial lower and upper bounds are determined by forecasted interest rate
			rL = (1.0d0-bspct)*rf
			rH = (1.0d0+bspct)*rf
			if (bsfixr==1) then
				wL = wf
				wH = wf
			else
				wL = (znow**(1.0d0/(1.0d0-ALPHA)))*(1.0d0-ALPHA)*(((rL+DELTA)/ALPHA)**(ALPHA/(ALPHA-1.0d0)))
				wH = (znow**(1.0d0/(1.0d0-ALPHA)))*(1.0d0-ALPHA)*(((rH+DELTA)/ALPHA)**(ALPHA/(ALPHA-1.0d0)))
            end if

		else

			wL = (1.0d0-bspct)*wf
			wH = (1.0d0+bspct)*wf
			if (bsfixr==1) then
				rL = rf
				rH = rf
			else
				rL = (znow**(1.0d0/ALPHA))*ALPHA*((wL/(1.0d0-ALPHA))**((ALPHA-1.0d0)/ALPHA)) - DELTA
				rH = (znow**(1.0d0/ALPHA))*ALPHA*((wH/(1.0d0-ALPHA))**((ALPHA-1.0d0)/ALPHA)) - DELTA
			end if

        end if
        

		call pricemap(agrid,kgrid,bgrid,xgrid,mgrid,invTa,cfmat0,EVmat0,mumat,idmat,iz,mp,wL,rL,ALOW,B0,HBAR,kappa1,kappa2,Yss,Trscale,Transprog1,Transprog2,T0, &
			pbij,pxij,tol,countmkt,calcma,linflag,hsvflag,Cagg,Hs,Taxagg,Transagg,LsL,Tmumat,cdist,hdist,taxdist,transdist,ldist,multhres)
        KLsL = Ks/LsL
        KLaggL = (wL/(znow*(1.0d0-ALPHA)))**(1.0d0/alpha)
  
		call pricemap(agrid,kgrid,bgrid,xgrid,mgrid,invTa,cfmat0,EVmat0,mumat,idmat,iz,mp,wH,rH,ALOW,B0,HBAR,kappa1,kappa2,Yss,Trscale,Transprog1,Transprog2,T0, &
			pbij,pxij,tol,countmkt,calcma,linflag,hsvflag,Cagg,Hs,Taxagg,Transagg,LsH,Tmumat,cdist,hdist,taxdist,transdist,ldist,multhres)
        KLsH = Ks/LsH
        KLaggH = (wH/(znow*(1.0d0-ALPHA)))**(1.0d0/alpha)
  
        if ((KLsL-KLaggL)*(KLsH-KLaggH) > 0.0d0) then

			print *, "  initial range is still not within the market-clearing bound!!"
			print *, "  thus, market-clearing is not imposed"
			countmkt = countmkt + 1
			call pricemap(agrid,kgrid,bgrid,xgrid,mgrid,invTa,cfmat0,EVmat0,mumat,idmat,iz,mp,wf,rf,ALOW,B0,HBAR,kappa1,kappa2,Yss,Trscale,Transprog1,Transprog2,T0, &
				pbij,pxij,tol,countmkt,calcma,linflag,hsvflag,Cagg,Hs,Taxagg,Transagg,Ls,Tmumat,cdist,hdist,taxdist,transdist,ldist,multhres)
			! CK assumption
			Lagg = Ls
            Kagg = mnow
            Ks = Kagg
			w0 = znow*(1.0d0-ALPHA)*(mnow/Lagg)**ALPHA
			r0 = znow*ALPHA*(mnow/Lagg)**(ALPHA-1.0d0) - DELTA

		else

			do while (distancemkt>tolmkt)

				if (bsctr==1) then
					r0 = (rL+rH)/2.0d0
					if (bsfixr==1) then
						w0 = wf
					else
						w0 = (znow**(1.0d0/(1.0d0-ALPHA)))*(1.0d0-ALPHA)*(((r0+DELTA)/ALPHA)**(ALPHA/(ALPHA-1.0d0)))
					end if
				else
					w0 = (wL+wH)/2.0d0
					if (bsfixr==1) then
						r0 = rf
					else
						r0 = (znow**(1.0d0/ALPHA))*ALPHA*((w0/(1.0d0-ALPHA))**((ALPHA-1.0d0)/ALPHA)) - DELTA
					end if
				end if

				! Ls = Household implied L
				countmkt = countmkt + 1
				call pricemap(agrid,kgrid,bgrid,xgrid,mgrid,invTa,cfmat0,EVmat0,mumat,idmat,iz,mp,w0,r0,ALOW,B0,HBAR,kappa1,kappa2,Yss,Trscale,Transprog1,Transprog2,T0, &
					pbij,pxij,tol,countmkt,calcma,linflag,hsvflag,Cagg,Hs,Taxagg,Transagg,Ls,Tmumat,cdist,hdist,taxdist,transdist,ldist,multhres)
                KLs = Ks/Ls
                KLagg = (w0/(znow*(1.0d0-ALPHA)))**(1.0d0/alpha)
                g0 = (KLs-KLagg)*(KLsL-KLaggL)

				if (bsctr==1) then

					if (g0>0.0d0) then
						rL = r0
					else
						rH = r0
					end if
					distancemkt = abs(log(rH)-log(rL))

				else

					if (g0>0.0d0) then
						wL = w0
					else
						wH = w0
					end if
					distancemkt = abs(log(wH)-log(wL))

				end if

				if (countmkt == maxcountmkt) exit
				
			end do ! end of bisection

			r0 = (znow**(1.0d0/ALPHA))*ALPHA*((w0/(1.0d0-ALPHA))**((ALPHA-1.0d0)/ALPHA)) - DELTA
            Lagg = Ls    ! By construction

		end if

	end if


end subroutine bisect



subroutine pricemap(agrid,kgrid,bgrid,xgrid,mgrid,invTa,cfmat0,EVmat0,mumat,idmat,iz,mp,w,r,ALOW,B0,HBAR,kappa1,kappa2,Yss,Trscale,Transprog1,Transprog2,T0, &
	pbij,pxij,tol,countmkt,calcma,linflag,hsvflag,Cagg,Hs,Taxagg,Transagg,Ls,Tmumat,cdist,hdist,taxdist,transdist,ldist,multhres)

	real(8), intent(in) :: agrid(:), kgrid(:), bgrid(:), xgrid(:), mgrid(:), invTa(:,:), cfmat0(:,:,:,:,:), EVmat0(:,:,:,:), mumat(:,:)
	integer, intent(in) :: idmat(:,:), iz, countmkt, calcma, linflag, hsvflag
	real(8), intent(in) :: mp, w, r, ALOW, B0, HBAR, kappa1, kappa2, Trscale, Transprog1, Transprog2, T0, Yss, pbij(:,:), pxij(:,:), tol
	real(8), intent(out) :: Cagg, Hs, Taxagg, Transagg, Ls, Tmumat(:,:), cdist(:,:), hdist(:,:), taxdist(:,:), transdist(:,:), ldist(:,:)
	integer, intent(out) :: multhres

	real(8), allocatable :: cfmat(:,:,:), EVmat(:,:), EV(:), wkmat(:,:), Wmat(:,:), Nmat(:,:), minasseta(:), minassetd(:), minasset(:)
	integer, allocatable :: jkmat(:,:)
	real(8) betanow, xnow, anow, Trw, Trn, awhigh, anhigh, vwnow, awp, vnnow, anp, ap, wk, atilde, TW, TN, diffma, wm, taxw, taxn
	integer na, nk, ne, nm, ra, rm, ia, ik, ie, ib, ix, je, jb, jx, jk, kk, jm


	na = size(agrid,1)
	nk = size(kgrid,1)
	ne = size(idmat,1)
	nm = size(mgrid,1)
	ra = na-2
	rm = nm-2
	allocate(EVmat(na,nm),EV(na),jkmat(nk,ne),wkmat(nk,ne),Wmat(na,ne),Nmat(na,ne),minasseta(ne),minassetd(ne),minasset(ne))
	if (linflag==2) then
		allocate(cfmat(4,ra+1,1))
	else
		allocate(cfmat(16,ra+1,rm+1))
	end if

    if (Trscale == 0.0d0) then
        ! for a below atilde, not-working is not feasible (consumption negative)
        atilde = (alow)/(1.0d0+r)
    else
        call transfers(0.0d0,0.0d0,r,w,0.0d0,Yss,Trscale,Transprog1,Transprog2,T0,Trn)
        atilde = (alow - Trn)/(1.0d0+r)
    end if


	if (calcma>0) then
		! ****************************************
		! threshold asset for work decision
		minasseta = 0.0d0 ! ascending order
		minassetd = 0.0d0 ! descending order

		!$omp parallel do default(shared) private(ie,ib,ix,betanow,xnow,ia,anow,awp,anp,TW,TN,EVmat,cfmat,jm,wm,EV)
		do ie = 1,ne

			ib = idmat(ie,1)
			ix = idmat(ie,2)
			betanow = bgrid(ib)
			xnow = xgrid(ix)

			! now we know ie
			EVmat = EVmat0(:,:,ie,iz)
			if (linflag==0) then
				cfmat = cfmat0(:,:,:,ie,iz)
			! elseif (linflag==1) then
			! 	EVmat = EVmat0(:,:,ie,iz)
			elseif (linflag==2) then
				call findiw(mp,mgrid,jm,wm)
				EV = wm*EVmat0(:,jm,ie,iz) + (1.0d0-wm)*EVmat0(:,jm+1,ie,iz)
				cfmat(:,:,1) = spfit(invTa,EV,ra,agrid)
			end if

			do ia = 1,na

				anow = agrid(ia)
				call calcwnval(xnow,anow,r,w,mp,Yss,atilde,HBAR,Trscale,Transprog1,Transprog2,T0,betanow,B0,kappa1,kappa2,ALOW,agrid,mgrid,cfmat,EVmat,tol,linflag,0,hsvflag,awp,anp,TW,TN)
				Wmat(ia,ie) = TW
				Nmat(ia,ie) = TN

			end do

		end do
		!$omp end parallel do

		!$omp parallel do default(shared) private(ie,ib,ix,betanow,xnow,EVmat,cfmat,jm,wm,EV)
		do ie = 1,ne

			ib = idmat(ie,1)
			ix = idmat(ie,2)
			betanow = bgrid(ib)
			xnow = xgrid(ix)

			! now we know ie
			EVmat = EVmat0(:,:,ie,iz)
			if (linflag==0) then
				cfmat = cfmat0(:,:,:,ie,iz)
			elseif (linflag==2) then
				call findiw(mp,mgrid,jm,wm)
				EV = wm*EVmat0(:,jm,ie,iz) + (1.0d0-wm)*EVmat0(:,jm+1,ie,iz)
				cfmat(:,:,1) = spfit(invTa,EV,ra,agrid)
			end if

			minasseta(ie) = bisectma(agrid,mgrid,cfmat,EVmat,Wmat(:,ie),Nmat(:,ie),xnow,r,w,mp,Yss,atilde,HBAR,Trscale,Transprog1,Transprog2,T0,betanow,B0,kappa1,kappa2,ALOW,tol,1,linflag,hsvflag)
			minassetd(ie) = bisectma(agrid,mgrid,cfmat,EVmat,Wmat(:,ie),Nmat(:,ie),xnow,r,w,mp,Yss,atilde,HBAR,Trscale,Transprog1,Transprog2,T0,betanow,B0,kappa1,kappa2,ALOW,tol,0,linflag,hsvflag)

			if (calcma==1) then
				minasset = minassetd
			elseif (calcma==2) then
				minasset = minasseta
            end if

		end do
		!$omp end parallel do

		if (sum(abs(minasseta-minassetd),1)>tol) then
            
			multhres = 1
		else
			multhres = 0
        end if

	end if


	hdist = 0.0d0
	cdist = 0.0d0
	taxdist = 0.0d0
	transdist = 0.0d0
	ldist = 0.0d0

	!$omp parallel do default(shared) private(ie,ib,ix,betanow,xnow,ik,anow,Trw,Trn,awhigh,anhigh,vwnow,awp,vnnow,anp,ap,jk,wk,kk,EVmat,cfmat,jm,wm,EV,taxw,taxn)
	do ik = 1,nk

		anow = kgrid(ik)

		do ie = 1,ne

			ib = idmat(ie,1)
			ix = idmat(ie,2)
			betanow = bgrid(ib)
			xnow = xgrid(ix)

			! now we know ie
			EVmat = EVmat0(:,:,ie,iz)
			if (linflag==0) then
				cfmat = cfmat0(:,:,:,ie,iz)
			elseif (linflag==2) then
				call findiw(mp,mgrid,jm,wm)
				EV = wm*EVmat0(:,jm,ie,iz) + (1.0d0-wm)*EVmat0(:,jm+1,ie,iz)
				cfmat(:,:,1) = spfit(invTa,EV,ra,agrid)
            end if

			! NOTE: about 80% of hshlds are active
			if (mumat(ik,ie)>0.0d0) then

				if (calcma>0) then
					! NOTE: if we know minasset beforehand, # of calling gss is nk*ne + 2*na*ne + (# in bisection) < 2*nk*ne
					if (anow >= minasset(ie)) then
						! Not work
						call transfers(xnow,anow,r,w,0.0d0,Yss,Trscale,Transprog1,Transprog2,T0,Trn)
                        call taxfunc(0.0d0,anow,r,Yss,kappa1,kappa2,hsvflag,taxn)
						anhigh = (1.0d0+r)*anow - taxn + Trn
						call gss(ALOW,anhigh,anp,tol/100.d0,0,linflag,anow,xnow,agrid,mgrid,cfmat,EVmat,mp,w,r,taxn,Trn,betanow,B0,Hbar,kappa2)
						ap = anp
						hdist(ik,ie) = 0.0d0
						cdist(ik,ie) = anhigh-ap
	                    taxdist(ik,ie) = taxn
	                    transdist(ik,ie) = Trn
					else
						! Work
						call transfers(xnow,anow,r,w,HBAR,Yss,Trscale,Transprog1,Transprog2,T0,Trw)
                        call taxfunc(w*xnow*HBAR,anow,r,Yss,kappa1,kappa2,hsvflag,taxw)
						awhigh = w*xnow*HBAR + (1.0d0+r)*anow - taxw + Trw
						call gss(ALOW,awhigh,awp,tol/100.d0,1,linflag,anow,xnow,agrid,mgrid,cfmat,EVmat,mp,w,r,taxw,Trw,betanow,B0,Hbar,kappa2)
						ap = awp
						hdist(ik,ie) = HBAR
						cdist(ik,ie) = awhigh-ap
	                    taxdist(ik,ie) = taxw
	                    transdist(ik,ie) = Trw
					end if

                else
					call calcwnval(xnow,anow,r,w,mp,Yss,atilde,HBAR,Trscale,Transprog1,Transprog2,T0,betanow,B0,kappa1,kappa2,ALOW,agrid,mgrid,cfmat,EVmat,tol,linflag,0,hsvflag,awp,anp,vwnow,vnnow)

					if (vwnow>vnnow) then
						call transfers(xnow,anow,r,w,HBAR,Yss,Trscale,Transprog1,Transprog2,T0,Trw)
                        call taxfunc(w*xnow*HBAR,anow,r,Yss,kappa1,kappa2,hsvflag,taxw)
						awhigh = w*xnow*HBAR + (1.0d0+r)*anow - taxw + Trw
						ap = awp
						hdist(ik,ie) = HBAR
						cdist(ik,ie) = awhigh-ap
	                    taxdist(ik,ie) = taxw
	                    transdist(ik,ie) = Trw
					else
						call transfers(xnow,anow,r,w,0.0d0,Yss,Trscale,Transprog1,Transprog2,T0,Trn)
                        call taxfunc(0.0d0,anow,r,Yss,kappa1,kappa2,hsvflag,taxn)
						anhigh = (1.0d0+r)*anow - taxn + Trn
						ap = anp
						hdist(ik,ie) = 0.0d0
						cdist(ik,ie) = anhigh-ap
	                    taxdist(ik,ie) = taxn
	                    transdist(ik,ie) = Trn
	                end if

				end if

                ldist(ik,ie) = xnow*hdist(ik,ie)

				! finding indices and weights
				! NOTE: finding indices and weights on grid points of k in the next periods
				call findjw(ap,kgrid,jk,wk)
				jkmat(ik,ie) = min(max(1,jk),nk-1)
				wkmat(ik,ie) = wk

			end if

		end do

	end do
	!$omp end parallel do

	! transition
	Tmumat = 0.0d0
	do ie = 1,ne

		ib = idmat(ie,1)
		ix = idmat(ie,2)

		do ik = 1,nk

			if (mumat(ik,ie) > 0.0d0) then

				jk = jkmat(ik,ie)
				wk = wkmat(ik,ie)

				do je = 1,ne

					jb = idmat(je,1)
					jx = idmat(je,2)
					Tmumat(jk,je) = Tmumat(jk,je) + pbij(ib,jb)*pxij(ix,jx)*wk*mumat(ik,ie)
					Tmumat(jk+1,je) = Tmumat(jk+1,je) + pbij(ib,jb)*pxij(ix,jx)*(1.0d0-wk)*mumat(ik,ie)

				end do

			end if

		end do

	end do

	Cagg = 0.0d0
	Hs = 0.0d0
	Taxagg = 0.0d0
	Transagg = 0.0d0
	Ls = 0.0d0

	do ie = 1,ne

		Cagg = Cagg + sum(mumat(:,ie)*cdist(:,ie),1)
		Hs   = Hs   + sum(mumat(:,ie)*hdist(:,ie),1)
		Taxagg = Taxagg + sum(mumat(:,ie)*taxdist(:,ie),1)
		Transagg = Transagg + sum(mumat(:,ie)*transdist(:,ie),1)
		Ls   = Ls   + sum(mumat(:,ie)*ldist(:,ie),1)

    end do    


end subroutine pricemap


function bisectma(agrid,mgrid,cfmat,EVmat,Wvalvec,Nvalvec,xnow,r,w,mp,Yagg,atilde,HBAR,Trscale,Transprog1,Transprog2,T0,betanow,B0,kappa1,kappa2,ALOW,tol,ascend,linflag,hsvflag) result(ma0)


	real(8), intent(in) :: agrid(:), mgrid(:), cfmat(:,:,:), EVmat(:,:), Wvalvec(:), Nvalvec(:), xnow, r, w, mp, Yagg, atilde, &
		HBAR, Trscale, Transprog1, Transprog2, T0, betanow, B0, kappa1, kappa2, ALOW, tol
	integer, intent(in) :: ascend, linflag, hsvflag
	real(8) ma0, malow, mahigh, WNdiffL, WNdiffH, WNdiff0, awp, anp, TW, TN, distance
	integer ia, na, iteration


	na = size(agrid,1)

	if (ascend) then
		! ascending order
		! NOTE: find minimum ia so that W(ia)<N(ia)
		do ia = 1,na
		    if ((Wvalvec(ia)-Nvalvec(ia)) < 0.0d0) exit
		end do

	else
		! descending order
		! NOTE: find minimum ia so that W(ia)>N(ia)
		do ia = na,1,-1
		    if ((Wvalvec(ia)-Nvalvec(ia)) > 0.0d0) exit
		end do
		ia = ia + 1

	end if

	if (ia == 1) then

	    ma0 = agrid(1)

	elseif (ia > na) then

	    ma0 = agrid(na)

	else

	    malow = agrid(ia-1)
	    WNdiffL = Wvalvec(ia-1) - Nvalvec(ia-1)
	    mahigh = agrid(ia)
	    WNdiffH = Wvalvec(ia) - Nvalvec(ia)

	    if (WNdiffL*WNdiffH > 0) then

		    print *, "  Bisection prerequisite violated!"
	        write(*,"('  ia-1 ', I4, '   ia ', I4)") ia-1, ia

		else

			! bisection
			distance = 1d+4
	        iteration = 0

	        do while (distance > tol*10.0d0)

	            ma0 = (malow+mahigh)/2.0d0

				call calcwnval(xnow,ma0,r,w,mp,Yagg,atilde,HBAR,Trscale,Transprog1,Transprog2,T0,betanow,B0,kappa1,kappa2,ALOW,agrid,mgrid,cfmat,EVmat,tol,linflag,0,hsvflag,awp,anp,TW,TN)

	            WNdiff0 = TW-TN

	            ! f(x0)*f(xl)
	            if (WNdiff0*WNdiffL > 0) then
	                malow = ma0
	            else
	                mahigh = ma0
	            end if

	            distance = mahigh-malow
	            iteration = iteration + 1

	        end do ! while loop for bisection

	    end if

	end if


end function bisectma


subroutine calcforecast(drop,izvec,zgrid,ymat,DELTA,kappakpnew,kappawnew,kapparnew,kappalnew,rsq,conzflag)

	integer, intent(in) :: drop, izvec(:), conzflag
	real(8), intent(in) :: ymat(:,:), DELTA, zgrid(:)
	real(8), intent(out) :: kappakpnew(:,:), kappawnew(:,:), kapparnew(:,:), kappalnew(:,:), rsq(:,:)
	integer iz, nz, simTT, num
    real(8), allocatable :: mvec(:), wvec(:), rvec(:), lvec(:), XX(:,:), yy(:,:), ee(:,:), yyhat(:,:), zvec(:)
    real(8) AA(2,2), bb(2,4), det, invAA(2,2), kappa(2,4), temp(1), yymean(4), AA0(3,3), bb0(3,4), invAA0(3,3), kappa0(3,4)
    real(8) AA3(4,4), bb3(4,4), invAA3(4,4), kappa3(4,4)   ! for option: conzflag == 3


    nz = size(kappakpnew,2)
	simTT = size(ymat,1)
	allocate(mvec(simTT),wvec(simTT),rvec(simTT),lvec(simTT),zvec(simTT))
	mvec = ymat(:,1)
	wvec = ymat(:,2)
    ! rvec is the rental rate
	rvec = ymat(:,3) + DELTA
    lvec = ymat(:,7)
	zvec = ymat(:,9)

	if (conzflag==1) then

        do iz = 1,nz

		    num = count(izvec(drop+1:simTT-1)==iz)

		    allocate(XX(num,2),yy(num,4),ee(num,4),yyhat(num,4))
		    XX(:,1) = 1.0d0
		    XX(:,2) = log(pack(mvec(drop+1:simTT-1),izvec(drop+1:simTT-1)==iz))     ! K
		    yy(:,1) = log(pack(mvec(drop+2:simTT),  izvec(drop+1:simTT-1)==iz))     ! K'
		    yy(:,2) = log(pack(wvec(drop+1:simTT-1),izvec(drop+1:simTT-1)==iz))
		    yy(:,3) = log(pack(rvec(drop+1:simTT-1),izvec(drop+1:simTT-1)==iz))
            yy(:,4) = log(pack(lvec(drop+1:simTT-1),izvec(drop+1:simTT-1)==iz))

            yymean = sum(yy,1)/num

		    AA = matmul(transpose(XX),XX)
		    bb = matmul(transpose(XX),yy)
            ! NOTE: using mkl-lapack
            invAA = inv(AA)
		    kappa = matmul(invAA,bb)

			kappakpnew(1,iz) = kappa(1,1)
			kappakpnew(2,iz) = kappa(2,1)
		    kappawnew(1,iz) = kappa(1,2)
			kappawnew(2,iz) = kappa(2,2)
		    kapparnew(1,iz) = kappa(1,3)
			kapparnew(2,iz) = kappa(2,3)
            kappalnew(1,iz) = kappa(1,4)
			kappalnew(2,iz) = kappa(2,4)

		    yyhat(:,1) = kappa(1,1) + kappa(2,1)*XX(:,2)
		    yyhat(:,2) = kappa(1,2) + kappa(2,2)*XX(:,2)
		    yyhat(:,3) = kappa(1,3) + kappa(2,3)*XX(:,2)
            yyhat(:,4) = kappa(1,4) + kappa(2,4)*XX(:,2)

			ee = yy - yyhat
		    temp = matmul(reshape(ee(:,1),(/1,num/)),ee(:,1))
		    temp = temp/matmul(reshape(yy(:,1)-yymean(1),(/1,num/)),yy(:,1)-yymean(1))
		    rsq(1,iz) = 1.0d0-temp(1)
		    temp = matmul(reshape(ee(:,2),(/1,num/)),ee(:,2))
		    temp = temp/matmul(reshape(yy(:,2)-yymean(2),(/1,num/)),yy(:,2)-yymean(2))
		    rsq(2,iz) = 1.0d0-temp(1)
		    temp = matmul(reshape(ee(:,3),(/1,num/)),ee(:,3))
		    temp = temp/matmul(reshape(yy(:,3)-yymean(3),(/1,num/)),yy(:,3)-yymean(3))
		    rsq(3,iz) = 1.0d0-temp(1)
            temp = matmul(reshape(ee(:,4),(/1,num/)),ee(:,4))
		    temp = temp/matmul(reshape(yy(:,4)-yymean(4),(/1,num/)),yy(:,4)-yymean(4))
		   
		    rsq(4,iz) = 1.0d0-temp(1)

		    deallocate(XX,yy,ee,yyhat)

        end do

    elseif (conzflag == 3) then

        num = simTT - drop - 1

        allocate(XX(num,4),yy(num,4),ee(num,4),yyhat(num,4))
		XX(:,1) = 1.0d0
		XX(:,2) = log(mvec(drop+1:simTT-1))
        XX(:,3) = log(zvec(drop+1:simTT-1))
        XX(:,4) = log(zvec(drop+1:simTT-1))*log(mvec(drop+1:simTT-1))
		yy(:,1) = log(mvec(drop+2:simTT))
		yy(:,2) = log(wvec(drop+1:simTT-1))
		yy(:,3) = log(rvec(drop+1:simTT-1))
        yy(:,4) = log(lvec(drop+1:simTT-1))


        yymean = sum(yy,1)/num

		AA3 = matmul(transpose(XX),XX)
		bb3 = matmul(transpose(XX),yy)
		! NOTE: using mkl-lapack
        invAA3 = inv(AA3)
		kappa3 = matmul(invAA3,bb3)

		kappakpnew(1,1) = kappa3(1,1)
        kappakpnew(2,1) = kappa3(2,1)
        kappakpnew(3,1) = kappa3(3,1)
        kappakpnew(4,1) = kappa3(4,1)
        kappawnew(1,1) = kappa3(1,2)
        kappawnew(2,1) = kappa3(2,2)
        kappawnew(3,1) = kappa3(3,2)
        kappawnew(4,1) = kappa3(4,2)
		kapparnew(1,1) = kappa3(1,3)
        kapparnew(2,1) = kappa3(2,3)
        kapparnew(3,1) = kappa3(3,3)
        kapparnew(4,1) = kappa3(4,3)
        kappalnew(1,1) = kappa3(1,4)
        kappalnew(2,1) = kappa3(2,4)
        kappalnew(3,1) = kappa3(3,4)
        kappalnew(4,1) = kappa3(4,4)

		yyhat(:,1) = kappa3(1,1) + kappa3(2,1)*XX(:,2) + kappa3(3,1)*XX(:,3) + kappa3(4,1)*XX(:,4)
		yyhat(:,2) = kappa3(1,2) + kappa3(2,2)*XX(:,2) + kappa3(3,2)*XX(:,3) + kappa3(4,2)*XX(:,4)
		yyhat(:,3) = kappa3(1,3) + kappa3(2,3)*XX(:,2) + kappa3(3,3)*XX(:,3) + kappa3(4,3)*XX(:,4)
        yyhat(:,4) = kappa3(1,4) + kappa3(2,4)*XX(:,2) + kappa3(3,4)*XX(:,3) + kappa3(4,4)*XX(:,4)
		ee = yy - yyhat
		temp = matmul(reshape(ee(:,1),(/1,num/)),ee(:,1))
		temp = temp/matmul(reshape(yy(:,1)-yymean(1),(/1,num/)),yy(:,1)-yymean(1))
		rsq(1,1) = 1.0d0-temp(1)
		temp = matmul(reshape(ee(:,2),(/1,num/)),ee(:,2))
		temp = temp/matmul(reshape(yy(:,2)-yymean(2),(/1,num/)),yy(:,2)-yymean(2))
		rsq(2,1) = 1.0d0-temp(1)
		temp = matmul(reshape(ee(:,3),(/1,num/)),ee(:,3))
		temp = temp/matmul(reshape(yy(:,3)-yymean(3),(/1,num/)),yy(:,3)-yymean(3))
		rsq(3,1) = 1.0d0-temp(1)
        temp = matmul(reshape(ee(:,4),(/1,num/)),ee(:,4))
		temp = temp/matmul(reshape(yy(:,4)-yymean(4),(/1,num/)),yy(:,4)-yymean(4))
		
		rsq(4,1) = 1.0d0-temp(1)

		deallocate(XX,yy,ee,yyhat)

	else

        num = simTT - drop - 1

        allocate(XX(num,3),yy(num,4),ee(num,4),yyhat(num,4))
		XX(:,1) = 1.0d0
		XX(:,2) = log(mvec(drop+1:simTT-1))
        XX(:,3) = log(zvec(drop+1:simTT-1))
		yy(:,1) = log(mvec(drop+2:simTT))
		yy(:,2) = log(wvec(drop+1:simTT-1))
		yy(:,3) = log(rvec(drop+1:simTT-1))
        yy(:,4) = log(lvec(drop+1:simTT-1))

        yymean = sum(yy,1)/num

		AA0 = matmul(transpose(XX),XX)
		bb0 = matmul(transpose(XX),yy)
		! NOTE: using mkl-lapack
        invAA0 = inv(AA0)
		kappa0 = matmul(invAA0,bb0)

		kappakpnew(1,1) = kappa0(1,1)
        kappakpnew(2,1) = kappa0(2,1)
        kappakpnew(3,1) = kappa0(3,1)
        kappawnew(1,1) = kappa0(1,2)
        kappawnew(2,1) = kappa0(2,2)
        kappawnew(3,1) = kappa0(3,2)
		kapparnew(1,1) = kappa0(1,3)
        kapparnew(2,1) = kappa0(2,3)
        kapparnew(3,1) = kappa0(3,3)
        kappalnew(1,1) = kappa0(1,4)
        kappalnew(2,1) = kappa0(2,4)
        kappalnew(3,1) = kappa0(3,4)

		yyhat(:,1) = kappa0(1,1) + kappa0(2,1)*XX(:,2) + kappa0(3,1)*XX(:,3)
		yyhat(:,2) = kappa0(1,2) + kappa0(2,2)*XX(:,2) + kappa0(3,2)*XX(:,3)
		yyhat(:,3) = kappa0(1,3) + kappa0(2,3)*XX(:,2) + kappa0(3,3)*XX(:,3)
        yyhat(:,4) = kappa0(1,4) + kappa0(2,4)*XX(:,2) + kappa0(3,4)*XX(:,3)
		ee = yy - yyhat
		temp = matmul(reshape(ee(:,1),(/1,num/)),ee(:,1))
		temp = temp/matmul(reshape(yy(:,1)-yymean(1),(/1,num/)),yy(:,1)-yymean(1))
		rsq(1,1) = 1.0d0-temp(1)
		temp = matmul(reshape(ee(:,2),(/1,num/)),ee(:,2))
		temp = temp/matmul(reshape(yy(:,2)-yymean(2),(/1,num/)),yy(:,2)-yymean(2))
		rsq(2,1) = 1.0d0-temp(1)
		temp = matmul(reshape(ee(:,3),(/1,num/)),ee(:,3))
		temp = temp/matmul(reshape(yy(:,3)-yymean(3),(/1,num/)),yy(:,3)-yymean(3))
		rsq(3,1) = 1.0d0-temp(1)
        temp = matmul(reshape(ee(:,4),(/1,num/)),ee(:,4))
		temp = temp/matmul(reshape(yy(:,4)-yymean(4),(/1,num/)),yy(:,4)-yymean(4))
		
		rsq(4,1) = 1.0d0-temp(1)

		deallocate(XX,yy,ee,yyhat)

	end if


end subroutine calcforecast


subroutine calcdisaggstat(HBAR,idmat,kgrid,mumat,cdist,hdist,taxdist,transdist,ldist, &
	mukx,consx,hx,taxx,transx,kx,lx,mukw,consw,hw,taxw,transw,kw,lw)


	integer, intent(in) :: idmat(:,:)
	real(8), intent(in) :: HBAR, kgrid(:), mumat(:,:), cdist(:,:), hdist(:,:), taxdist(:,:), transdist(:,:), ldist(:,:)
	real(8), intent(out) :: mukx(:), consx(:), hx(:), taxx(:), transx(:), kx(:), lx(:)
	real(8), intent(out) :: mukw(:), consw(:), hw(:), taxw(:), transw(:), kw(:), lw(:)
	integer ik, ie, nk, ne, ib, ix
	real(8) vep1, W5threshold(4)
	real(8), allocatable :: muk(:)


	nk = size(kgrid,1)
	ne = size(idmat,1)
	allocate(muk(nk))

	! calulcate disaggregated time series by productivity x
	mukx = 0.0d0
	consx = 0.0d0
	hx = 0.0d0
	taxx = 0.0d0
	transx = 0.0d0
	kx = 0.0d0
	lx = 0.0d0

	do ik = 1,nk

		do ie = 1,ne

			ib = idmat(ie,1)
			ix = idmat(ie,2)

			! print *, ix
			mukx(ix) = mukx(ix) + mumat(ik,ie)
			consx(ix) = consx(ix) + cdist(ik,ie)*mumat(ik,ie)
			hx(ix) = hx(ix) + (hdist(ik,ie)/HBAR)*mumat(ik,ie)
			taxx(ix) = taxx(ix) + taxdist(ik,ie)*mumat(ik,ie)
			transx(ix) = transx(ix) + transdist(ik,ie)*mumat(ik,ie)
			kx(ix) = kx(ix) + kgrid(ik)*mumat(ik,ie)
			lx(ix) = lx(ix) + ldist(ik,ie)*mumat(ik,ie)

		end do

	end do

	! calculate (time-varying) wealth threshold
	muk = sum(mumat,2)
	W5threshold = 0         ! index
	vep1 = 0.0d0
	do ik = 1,nk
		vep1 = vep1 + muk(ik)
		if (vep1 >= 0.2d0 .and. W5threshold(1) == 0) then
			W5threshold(1) = ik
		elseif (vep1 >= 0.4d0 .and. W5threshold(2) == 0) then
			W5threshold(2) = ik
		elseif (vep1 >= 0.6d0 .and. W5threshold(3) == 0) then
			W5threshold(3) = ik
		elseif (vep1 >= 0.8d0 .and. W5threshold(4) == 0) then
			W5threshold(4) = ik
		end if
	end do

	! calulcate disaggregated time series by wealth w
	mukw = 0.0d0
	consw = 0.0d0
	hw = 0.0d0
	taxw = 0.0d0
	transw = 0.0d0
	kw = 0.0d0
	lw = 0.0d0

	! mumat, cdist, hdist, taxdist, transdist, ldist
	do ik = 1,nk

		do ie = 1,ne

			ib = idmat(ie,1)
			ix = idmat(ie,2)

			if (ik <= w5threshold(1)) then
				mukw(1) = mukw(1) + mumat(ik,ie)
				consw(1) = consw(1) + cdist(ik,ie)*mumat(ik,ie)
				hw(1) = hw(1) + (hdist(ik,ie)/HBAR)*mumat(ik,ie)
				taxw(1) = taxw(1) + taxdist(ik,ie)*mumat(ik,ie)
				transw(1) = transw(1) + transdist(ik,ie)*mumat(ik,ie)
				kw(1) = kw(1) + kgrid(ik)*mumat(ik,ie)
				lw(1) = lw(1) + ldist(ik,ie)*mumat(ik,ie)
			else if (ik > w5threshold(1) .and. (ik <= w5threshold(2))) then
				mukw(2) = mukw(2) + mumat(ik,ie)
				consw(2) = consw(2) + cdist(ik,ie)*mumat(ik,ie)
				hw(2) = hw(2) + (hdist(ik,ie)/HBAR)*mumat(ik,ie)
				taxw(2) = taxx(2) + taxdist(ik,ie)*mumat(ik,ie)
				transw(2) = transx(2) + transdist(ik,ie)*mumat(ik,ie)
				kw(2) = kw(2) + kgrid(ik)*mumat(ik,ie)
				lw(2) = lw(2) + ldist(ik,ie)*mumat(ik,ie)
			else if (ik > w5threshold(2) .and. (ik <= w5threshold(3))) then
				mukw(3) = mukw(3) + mumat(ik,ie)
				consw(3) = consw(3) + cdist(ik,ie)*mumat(ik,ie)
				hw(3) = hw(3) + (hdist(ik,ie)/HBAR)*mumat(ik,ie)
				taxw(3) = taxw(3) + taxdist(ik,ie)*mumat(ik,ie)
				transw(3) = transw(3) + transdist(ik,ie)*mumat(ik,ie)
				kw(3) = kw(3) + kgrid(ik)*mumat(ik,ie)
				lw(3) = lw(3) + ldist(ik,ie)*mumat(ik,ie)
			else if (ik > w5threshold(3) .and. (ik <= w5threshold(4))) then
				mukw(4) = mukw(4) + mumat(ik,ie)
				consw(4) = consw(4) + cdist(ik,ie)*mumat(ik,ie)
				hw(4) = hw(4) + (hdist(ik,ie)/HBAR)*mumat(ik,ie)
				taxw(4) = taxw(4) + taxdist(ik,ie)*mumat(ik,ie)
				transw(4) = transw(4) + transdist(ik,ie)*mumat(ik,ie)
				kw(4) = kw(4) + kgrid(ik)*mumat(ik,ie)
				lw(4) = lw(4) + ldist(ik,ie)*mumat(ik,ie)
			else if (ik > w5threshold(4)) then
				mukw(5) = mukw(5) + mumat(ik,ie)
				consw(5) = consw(5) + cdist(ik,ie)*mumat(ik,ie)
				hw(5) = hw(5) + (hdist(ik,ie)/HBAR)*mumat(ik,ie)
				taxw(5) = taxw(5) + taxdist(ik,ie)*mumat(ik,ie)
				transw(5) = transw(5) + transdist(ik,ie)*mumat(ik,ie)
				kw(5) = kw(5) + kgrid(ik)*mumat(ik,ie)
				lw(5) = lw(5) + ldist(ik,ie)*mumat(ik,ie)
			end if

		end do

	end do


end subroutine calcdisaggstat


end module mod_outer
