module mod_policies

use mod_parameters, only: T1amin, Transprog2, tauss

implicit none


contains

subroutine taxfunc(laborinc,anow,r,Y,kappa1,kappa2,hsvflag,tax)

    real(8), intent(in) :: laborinc, anow, r, Y, kappa1, kappa2
	integer, intent(in) :: hsvflag
    real(8), intent(out) :: tax
    real(8) :: avgtaxrate, income, meanearn
    
    meanearn = (1.0d0-0.36d0)*Y
    
	if (hsvflag==1) then

        if (laborinc == 0.0d0) then
            avgtaxrate = 0.0d0
        else
            
            avgtaxrate = max(1.0d0 - kappa2*((laborinc/meanearn)**(-kappa1)),0.0d0) + tauss
               
        end if
        
        tax = avgtaxrate*laborinc
        
	else

		tax = tauss*laborinc

	end if

end subroutine taxfunc


subroutine transfers(xnow,anow,interest,wage,work,output,Trscale,Transprog1,Transprog2,T0,trans)

    real(8), intent(in) :: xnow, anow, interest, wage, work, output, Trscale, Transprog1, Transprog2, T0
    real(8), intent(out) :: trans
    real(8) :: income, assetbar, laborinc, eitc
    
    
    ! weighted income
    laborinc = wage*xnow*work    
    
    income = laborinc + interest*max(anow,0.0d0)    
    
    trans = T0 + Trscale*((1.0d0+(income))**(-Transprog1))*((1.0d0+(max(anow,0.0d0)))**(-Transprog2))
    

end subroutine transfers


subroutine sortdist(nsize, gridold, muold, gridold2, gridnew, munew, gridnew2)


    integer, intent(in) :: nsize
    real(8), intent(in) :: gridold(nsize), muold(nsize), gridold2(nsize)
    real(8), intent(out) :: gridnew(nsize), munew(nsize), gridnew2(nsize)
    integer :: irow, krow
    real(8) :: gridtemp, mutemp, gridtemp2

    gridnew = gridold
    munew = muold
    gridnew2 = gridold2

    do irow = 1,nsize
        krow = minloc( gridnew(irow:nsize), dim=1 ) + irow - 1
        ! gridold is the sorting basis

        gridtemp = gridnew(irow)
        mutemp = munew(irow)
        gridtemp2 = gridnew2(irow)

        gridnew(irow) = gridnew(krow)
        munew(irow) = munew(krow)
        gridnew2(irow) = gridnew2(krow)

        gridnew(krow) = gridtemp
        munew(krow) = mutemp
        gridnew2(krow) = gridtemp2

    end do


end subroutine sortdist


end module mod_policies
