! Main program for "Implications of Government Transfers for Aggregate Labor Market Fluctuations"
! by Youngsoo Jang, Takeki Sunakawa and Minchul Yum

program solve


	use mod_parameters
    use mod_utils
	use mod_calcss
    use mod_inner
    use mod_outer
	implicit none


    real(8) agrid(na), mgrid(nm), xgrid(nx), zgrid(nz), pxij(nx,nx), pzij(nz,nz), xss(nx), ALOW
    real(8) bgrid(nb), pbij(nb,nb)
    integer idmat(ne,2), error
	integer ia, ik, im, ib, ix, jx, ie, iz, jz, iterout, iteroutvec(100), i, foreind, izvec(simTT), time, iv
    real(8) w, r, Lss, Kss, Yss, TY, TrY, Trss, kgrid(nk), mumat(nk,ne), muk(nk), mustationary(nk,ne), ndist(nk,ne), Ks, Ls, EPratio, wss, rss
    real(8) TrInc5(5), TrW5(5), EmpW5(5), stdlogwage, stdlogearn, rhoxest, Tr12, Tr1, Tr2, Tr45

    real(8) Vmat(na,nm,ne,nz)  
    real(8) apmat(na,nm,ne,nz), awpmat(na,nm,ne,nz), anpmat(na,nm,ne,nz), lmat(na,nm,ne,nz)
    real(8) aggsim(simTT,10), disaggsim(simTT,nx,6)
    real(8), allocatable :: kappakp(:,:), kappaw(:,:), kappar(:,:), kappal(:,:), kappakpnew(:,:), kappawnew(:,:), kapparnew(:,:), kappalnew(:,:), rsquared(:,:)
    real(8) diffkp, diffw, diffr, diffl, diffout, damp, diffkpvec(100), diffwvec(100), diffrvec(100), difflvec(100), bspct, Vvec(na*nm*ne*nz)
    real(8) xghz(nghz), wghz(nghz), zvec(simTT)

    integer, parameter :: ra = na-2
    integer, parameter :: rm = nm-2
    real(8) invTa(ra,ra), invTm(rm,rm)

    integer cto1, cto2, cro
    real(8) eptimeo
	! vsl
	integer, parameter :: method = VSL_RNG_METHOD_UNIFORM_STD
    integer errcode, brng, seed, maxsimind
    type(vsl_stream_state) :: stream
	real(8), allocatable :: murand(:), xrand(:)


    call system_clock(cto1, cro)

    ! index for individual exogenous state
	do ib = 1,nb

		do ix = 1,nx

			ie = (ib-1)*nx + ix
			idmat(ie,1) = ib
			idmat(ie,2) = ix

		end do

	end do

	maxsimind = nint(0.5d0*NS) !nint((maxval(xss)*1.1d0)*NS)
	! for values of brng parameter, see https://software.intel.com/en-us/mkl-developer-reference-fortran-brng-parameter-definition#TBL10-2
	brng = vsl_brng_sfmt19937 ! use Mersenne Twister
	seed = 092017

	errcode = vslnewstream(stream,brng,seed)
	allocate(murand(NS),xrand(nx*maxsimind))
	errcode = vdrnguniform(method,stream,NS,murand,0.0d0,1.0d0)
	errcode = vdrnguniform(method,stream,nx*maxsimind,xrand,0.0d0,1.0d0)

    if (ssini0==1) then
        call readss(idmat,ALPHA,BETA,agrid,kgrid,bgrid,xgrid,pbij,pxij,xss,mumat,ndist,muk,Kss,Lss,Yss,ALOW)
    else
        write(*,*) "calculating the steady state..."
		call calcss(idmat,murand,xrand,agrid,kgrid,bgrid,xgrid,pbij,pxij,xss,ALOW, &
			w,r,Lss,Kss,Yss,TY,TrY,Trss,mumat,ndist,Ks,Ls,EPratio,TrInc5,TrW5,EmpW5,error, &
			ALPHA,DELTA,HBAR,phi,B0,BETA,RHOX,SDINOVX,kappa1,kappa2,Trscale,Transprog1,Transprog2,T0, &
			tol,tolss,toldist,maxiter,dampss0,congrid,deggrida,deggridk,1,diagnum,logdiff,tauflag,calcma,linflag,howflag,hsvflag,mx)
    end if
    wss = (1.0d0-ALPHA)*Yss/Lss
    rss = ALPHA*Yss/Kss

    ! store the staionary distribution
    mustationary = mumat

    ! set up the grid points
    mgrid = linspace((1.0d0-mpct)*Kss,(1.0d0+mpct)*Kss,nm)

    ! grids for TFP
    call qnwnorm(nghz,0.0d0,SDINOVZ,xghz,wghz)
    if (conzflag==2) then
        zgrid = linspace(-mz*SDINOVZ/sqrt(1.0d0-RHOZ**2),mz*SDINOVZ/sqrt(1.0d0-RHOZ**2),nz)
    else
        if (tauflagagg==1) then
            call tauchen(nz,0.0d0,RHOZ,SDINOVZ,mz,zgrid,pzij)
        else
            call rouwenhorst(RHOZ,SDINOVZ,nz,zgrid,pzij)
        end if
    end if

    zgrid = exp(zgrid)

    open(510, file="mgrid.txt")
    do im = 1,nm
        write(510,*) mgrid(im)
    end do
    close(510)

    open(520, file="zgrid.txt")
    do iz = 1,nz
        write(520,*) zgrid(iz)
    end do
    close(520)

    open(530, file="pzij.txt")
    do jz = 1,nz
        do iz = 1,nz
            write(530,*) pzij(iz,jz)
        end do
    end do
    close(530)

    ! setup splines
    invTa = spbas(ra,agrid)
    invTm = spbas(rm,mgrid)
    ! take the inverse of the basis function matrix
    invTa = inv(invTa)
    invTm = inv(invTm)

    !***************************************!
    !******* KRUSELL-SMITH ALGORITHM *******!
    !***************************************!
    ! generate random numbers
    izvec = genizvec(simTT,pzij)
    zvec = genzvec(simTT,RHOZ,SDINOVZ)

    ! initial value function and forecasting rules
    Vmat = 0.0d0

    if (conzflag==1) then

        allocate(kappakp(2,nz),kappaw(2,nz),kappar(2,nz),kappal(2,nz),kappakpnew(2,nz),kappawnew(2,nz),kapparnew(2,nz),kappalnew(2,nz),rsquared(4,nz))

        do iz = 1,nz
            ! NOTE: The initial forecasting rules are from log-linearized representative-agent model
            ! First run with fcsteqn = 1 and outermkt = 0, then use the converged forecasting rule for r (and mp and w as well) by fcsteqn = 3
            ! Note: 20191206: The below (original one) tend to work better
            kappakp(1,iz) = (1.0d0-0.9431d0)*log(Kss) + 0.1359d0*log(zgrid(iz))
            kappakp(2,iz) = 0.9431d0
            kappaw(1,iz) = log(wss)-0.5714d0*log(Kss) + 0.3699d0*log(zgrid(iz))
            kappaw(2,iz) = 0.5714d0
		
        end do

    elseif (conzflag==3) then

        allocate(kappakp(4,1),kappaw(4,1),kappar(4,1),kappal(4,1),kappakpnew(4,1),kappawnew(4,1),kapparnew(4,1),kappalnew(4,1),rsquared(4,1))

        kappakp(1,1) = 0.113d0
        kappakp(2,1) = 0.954d0
        kappakp(3,1) = 0.100d0
        kappakp(4,1) = 0.0d0
        kappaw(1,1) = -0.398d0
        kappaw(2,1) = 0.522d0
        kappaw(3,1) = 0.800d0
        kappaw(4,1) = 0.0d0
        kappar(1,1) = -1.394d0
        kappar(2,1) = -0.799d0
        kappar(3,1) = 1.356d0
        kappar(4,1) = 0.0d0
        kappal(1,1) = -0.31d0
        kappal(2,1) = -0.799d0
        kappal(3,1) = 1.356d0
        kappal(4,1) = 0.0d0

    else

        allocate(kappakp(3,1),kappaw(3,1),kappar(3,1),kappal(3,1),kappakpnew(3,1),kappawnew(3,1),kapparnew(3,1),kappalnew(3,1),rsquared(4,1))

        kappakp(1,1) = 0.113d0
        kappakp(2,1) = 0.954d0
        kappakp(3,1) = 0.100d0
        kappaw(1,1) = -0.398d0
        kappaw(2,1) = 0.522d0
        kappaw(3,1) = 0.800d0
        kappar(1,1) = -1.394d0
        kappar(2,1) = -0.799d0
        kappar(3,1) = 1.356d0
        kappal(1,1) = -0.31d0
        kappal(2,1) = -0.799d0
        kappal(3,1) = 1.356d0

    end if

    ! loading previous results
    if (vmatini0==1) call readwriteVmat(Vmat,1)
    if (fcstini0==1) call readwritekappa(kappakp,kappaw,kappar,kappal,ckappakp,ckappaw,ckappar,ckappal,conzflag,1) ! read


    !******* MAIN LOOP *******!
    diffout = 1d+4
    iterout = 1
    damp = dampout0
    bspct = bspct0

    if (outermkt==1) foreind = 1  ! =1 flag for market clearing

	do

  
            call inner(kappakp,kappaw,kappar,agrid,mgrid,invTa,invTm,zgrid,pzij,bgrid,xgrid,pbij,pxij,idmat,Yss, &
				Vmat,apmat,awpmat,anpmat,lmat,xghz,wghz, &
                ALPHA,DELTA,HBAR,ALOW,B0,BETA,RHOZ,kappa1,kappa2,Trscale,Transprog1,Transprog2,T0, &
                tol,maxiter,diagnum,logdiff,vmatini,iterout,fcsteqn,linflag,howflag,hsvflag,conzflag)

            call readwriteVmat(Vmat,0) 
       
        call outer(Vmat,kappakp,kappaw,kappar,agrid,mgrid,kgrid,invTa,invTm,zgrid,pzij,bgrid,xgrid,pbij,pxij,idmat,izvec,simTT,Yss, &
			aggsim,disaggsim,mumat,zvec,xghz,wghz, &
			ALPHA,DELTA,HBAR,ALOW,B0,BETA,RHOZ,kappa1,kappa2,Trscale,Transprog1,Transprog2,T0, &
            tol,tolmkt,outermkt,foreind,bsctr,bspct,maxcountmkt,calcma,bsfixr,diagnumout,fcsteqn,linflag,hsvflag,conzflag)

        call calcforecast(drop,izvec,zgrid,aggsim,DELTA,kappakpnew,kappawnew,kapparnew,kappalnew,rsquared,conzflag)

        diffkp = maxval(maxval(abs(kappakpnew-kappakp),1),1)
        diffw = maxval(maxval(abs(kappawnew-kappaw),1),1)
        diffr = maxval(maxval(abs(kapparnew-kappar),1),1)
        diffl = maxval(maxval(abs(kappalnew-kappal),1),1)

      
        diffout = max(max(diffkp,diffw),diffr)
     

        ! Once the forecasting rules are close to the ones in equiilibrium, the market-clearing conditions are strictly imposed.
        if ((outermkt==1) .and. (diffr<tolout*100.0d0)) then
            foreind = 1
        end if

        iteroutvec(iterout) = iterout
        diffkpvec(iterout)  = diffkp
        diffwvec(iterout)   = diffw
        diffrvec(iterout)   = diffr
        difflvec(iterout)   = diffl

        write(*,*) " "
        do i = 1,iterout
            write(*,"('  iteration ', I4, '   ||Tkappa-kappa|| = (', F8.5, ', ', F8.5, ', ', F8.5, ', ', F8.5, ')')") &
            iteroutvec(i), diffkpvec(i), diffwvec(i), diffrvec(i), difflvec(i)
        end do

        if (conzflag==1) then

            do iz = 1,nz
                write(*,"('  z = ', F8.5, ': R2 of K = ', F8.5, ', R2 of w = ', F8.5, ', R2 of r = ', F8.5, ', R2 of L = ', F8.5)") &
                zgrid(iz), rsquared(1,iz), rsquared(2,iz), rsquared(3,iz), rsquared(4,iz)
            end do
            write(*,*) " "

        else

            write(*,"('  R2 of K = ', F8.5, ', R2 of w = ', F8.5, ', R2 of r = ', F8.5, ', R2 of L = ', F8.5)") &
                rsquared(1,1), rsquared(2,1), rsquared(3,1), rsquared(4,1)
            write(*,*) " "

        end if

        ! store the simulated time series and forecasting rules
        open(300, file="izvec.txt")
        open(310, file="mvec.txt")
        open(320, file="wvec.txt")
        open(330, file="rvec.txt")
        open(340, file="yvec.txt")
        open(350, file="cvec.txt")
        open(360, file="xvec.txt")
        open(370, file="lvec.txt")
        open(380, file="hvec.txt")
        open(390, file="zvec.txt")
        open(400, file="trvec.txt")

        do time = 1,simT

            write(300,*) izvec(time+drop)
            write(310,*) aggsim(time+drop,1)
            write(320,*) aggsim(time+drop,2)
            write(330,*) aggsim(time+drop,3)
            write(340,*) aggsim(time+drop,4)
            write(350,*) aggsim(time+drop,5)
            write(360,*) aggsim(time+drop,6)
            write(370,*) aggsim(time+drop,7)
            write(380,*) aggsim(time+drop,8)
            write(390,*) aggsim(time+drop,9)
            write(400,*) aggsim(time+drop,10)

        end do

        close(300)
        close(310)
        close(320)
        close(330)
        close(340)
        close(350)
        close(360)
        close(370)
        close(380)
        close(390)
        close(400)

		if ((.NOT. diffout>tolout).or.(iterout>30)) exit

        kappakp = (1.0d0-damp)*kappakpnew + damp*kappakp
        kappaw  = (1.0d0-damp)*kappawnew  + damp*kappaw
        kappar  = (1.0d0-damp)*kapparnew  + damp*kappar
        kappal  = (1.0d0-damp)*kappalnew  + damp*kappal

		call readwritekappa(kappakp,kappaw,kappar,kappal,ckappakp,ckappaw,ckappar,ckappal,conzflag,0) ! write

		iterout = iterout + 1

    end do


    call system_clock(cto2, cro)
    eptimeo = (dble(cto2-cto1)/dble(cro))/60.0d0
    write(*,"('  Total Elasped time (minutes) = ', F10.2)") eptimeo


end program solve
