clear
use Panel_1969-2015_with_unemp, replace


global x = 5     // 5 or 10 or 15
global posonly = 0    // 1 positive change only
global negonly = 0     // 1 negative change only

global fthours = 1000
global unempexc = 1    // 1 exclude (default option)

// SRC(1000-3000) SEO(5000-7000) and 1997(3000-5000) new entry SEO also
keep if famid<3000
// Trim non-responsor
drop if htotearns==. 
replace htotearns=0 if htotearns<0

// drop top code of earnings in 1994 and 1995
drop if htotearns==9999999 &(year==1994 | year==1995)

// Calculate hourly wage (htotearns/wrkhrs)
gen hrlywage = htotearns/wrkhrs

// Infaltion adjustment with CPI-U (base year=2000)
// Note that the reference year is a year before the survey year
gen prlevel = 0
replace prlevel = 4.9483 if year == 1969
replace prlevel = 4.6921 if year == 1970
replace prlevel = 4.4381 if year == 1971
replace prlevel = 4.2519 if year == 1972
replace prlevel = 4.1196 if year == 1973
replace prlevel = 3.8784 if year == 1974
replace prlevel = 3.4929 if year == 1975
replace prlevel = 3.2007 if year == 1976
replace prlevel = 3.0264 if year == 1977
replace prlevel = 2.8416 if year == 1978
replace prlevel = 2.6411 if year == 1979
replace prlevel = 2.3719 if year == 1980
replace prlevel = 2.0898 if year == 1981
replace prlevel = 1.8944 if year == 1982
replace prlevel = 1.7845 if year == 1983
replace prlevel = 1.7289 if year == 1984
replace prlevel = 1.6574 if year == 1985
replace prlevel = 1.6004 if year == 1986
replace prlevel = 1.5712 if year == 1987
replace prlevel = 1.5158 if year == 1988
replace prlevel = 1.4556 if year == 1989
replace prlevel = 1.3887 if year == 1990
replace prlevel = 1.3175 if year == 1991
replace prlevel = 1.2643 if year == 1992
replace prlevel = 1.2274 if year == 1993
replace prlevel = 1.1917 if year == 1994
replace prlevel = 1.1619 if year == 1995
replace prlevel = 1.1299 if year == 1996
replace prlevel = 1.0975 if year == 1997
replace prlevel = 1.0729 if year == 1998
replace prlevel = 1.0564 if year == 1999
replace prlevel = 1.0336 if year == 2000
replace prlevel = 1 if year == 2001
replace prlevel = 0.9723 if year == 2002
replace prlevel = 0.9572 if year == 2003
replace prlevel = 0.9359 if year == 2004
replace prlevel = 0.9116 if year == 2005
replace prlevel = 0.8817 if year == 2006
replace prlevel = 0.8542 if year == 2007
replace prlevel = 0.8307 if year == 2008
replace prlevel = 0.7998 if year == 2009
replace prlevel = 0.8028 if year == 2010
replace prlevel = 0.7895 if year == 2011
replace prlevel = 0.7657 if year == 2012
replace prlevel = 0.75 if year == 2013
replace prlevel = 0.7391 if year == 2014
replace prlevel = 0.7275 if year == 2015

* SAMPLE RESTRICTION

* household heads only
keep if headrel==1 | headrel==10

* age repair
sort id year
by id: gen test=_n
by id: gen yeargap=year-year[_n-1]
by id: gen agegap=age-age[_n-1]
by id: replace age=age[_n-1]+yeargap if agegap!=yeargap
drop agegap

* age
keep if age>=22 & age<=64

* self-employed
drop if selfemp == 3

gen fullempst=0
replace fullempst=1 if wrkhrs >= $fthours
sum fullempst

// Adjust the inflation on earnings with CPI-U (base year= 2000)
replace htotearns = htotearns * prlevel
replace hrlywage = hrlywage * prlevel
replace networth = networth * prlevel


gen wagegroup=0

* Generating quintiles based on base year
forvalue i=1969/1997{
	quietly egen wage5_`i'=xtile(hrlywage) if hrlywage>0 & year==`i', n(5)
}

sort year id
table year, c(m hrlywage)




*********************************************


// Year Grouping
forvalue i=1969/1997{
	quietly gen yrlgrp`i' = 0
	quietly gen fullempst`i' = 1 if fullempst==1 & year==`i'
	* range 0 to x-1 => x year band
	quietly replace yrlgrp`i'=1 if year-`i'<= $x - 1 & year-`i'>=0 
}

* Definition of numobs'year' : maximum number of consecutive observations by id
forvalue i=1969/1997{
	sort id yrlgrp`i'
	by id yrlgrp`i': egen numobs`i'=sum(yrlgrp`i')	
}

sort id year


* Construct switch variable for each base year
* year range up to 1997-x+1
global yearmax = 1997- $x + 1

forvalue i=1969/$yearmax{	
	* numobs == x
	by id: gen swlabor`i'=fullempst-fullempst[_n-1] if numobs`i'==$x	
	
	if $unempexc == 1{
	quietly replace swlabor`i'=. if swlabor`i'== -1 & unempspell > 0 & unempspell ~=.
	}
	
	if $posonly == 1{
	    quietly replace swlabor`i'=0 if swlabor`i'== -1
	}
	if $negonly == 1{
	    quietly replace swlabor`i'=0 if swlabor`i'== 1
	}
	
	quietly replace swlabor`i'=1 if swlabor`i'== -1
	
	quietly replace swlabor`i'=. if year ==`i'	
	
}


sort id year
* Main statistics
* year range up to 1997-x+1
forvalue i=1969/$yearmax{
	* numobs == x
	by id: egen avgsw`i' = mean(swlabor`i') if numobs`i'==$x
}	
* year range up to 1997-x+1
forvalue i=1969/$yearmax{	
	* numobs == x
	su age hrlywage sex wage5_`i' swlabor`i' if numobs`i'==$x	
	* numobs == x
	table wage5_`i' if numobs`i'==$x, c(m avgsw`i') 
}	

