module mod_parameters
   
    
    integer, parameter :: outermkt = 1      ! =1 market clearing in outerloop
    integer, parameter :: ssini0 = 0        ! =1 reading the previous steady state results
    integer, parameter :: vmatini0 = 0      ! =1 reading the previous inner loop results saved in Vmat.txt for the first iteration
    integer, parameter :: fcstini0 = 1      ! =1 reading the previous forecasting rules saved in the files below
    integer, parameter :: na = 400      
    integer, parameter :: nk = 4000         ! for distributions

    integer, parameter :: linflag = 1       ! =0 use 2 dim spline over (a,m),
                                            ! =1 use linear interpolation over (a,m),
                                            ! =2 use 1 dim spline over a and linear interpolation over m
    integer, parameter :: calcma = 0        ! =1 use minasset, the threshold below which hshlds choose to work

    ! 0: optimized, _ini: initial value to be updated, ?: flag for initial value
    !******* metaparameters *******!
    real(8), parameter :: tol = 1d-4        ! tolerance level for innerloop and (tol/100) for optimization (NOTE: tol**2 is too strict)
    real(8), parameter :: tolss = 1d-4      ! tolerance level for market clearing in steady state
    real(8), parameter :: toldist = 1d-7    ! tolerance level for distribution (for stationary equilibrium)
    real(8), parameter :: tolout = 1d-3     ! tolerance level for outer loop
                                            ! NOTE: 1d-4 lead to divergence
    real(8), parameter :: tolmkt = 1d-5     ! tolerance level for market clearing in outerloop (NOTE: 1d-6 is too strict; 1d-4 is okay but leads to a bit changes)
    real(8), parameter :: dampss0 = 0.9d0   ! dampening weight for market clearing in stationary equilibrium (for old)
                                            ! NOTE: values close to 0.5 (e.g., 0.7) lead to divergence
    real(8), parameter :: dampout0 = 0.5d0  ! dampening weight for outer loop (for old)
    integer, parameter :: maxiter = 2000    ! the maximum number of value function iterations in innerloop
    integer, parameter :: diagnum = 1000    ! display result in every diagnum interval (innerloop)
    integer, parameter :: diagnumout = 2000 ! display result in every diagnum interval (outerloop)

    integer, parameter :: howflag = 1       ! =1 use Howard's improvement algorithm in innerloop
    integer, parameter :: conzflag = 0      ! =3 use continuous z (linear interpolation) and ln(z) & ln(z)*ln(K) as regressors in forecasting rules
                                            ! =2 use continuous z (linear interpolation) and ln(z)-as a regressor in forecasting rules
                                            ! =1 use z-conditional forecasting rules
                                            ! =0 use ln(z) as a regressor in forecasting rules

    integer, parameter :: maxcountmkt = 30  ! the maximum number of bisection iterations
    integer, parameter :: bsctr = 0         ! =1 bisection for market clearing over r, =0 over w
    integer, parameter :: bsfixr = 0        ! =1 fix r (w if bsctr=1) while updating w in market clearing
    real(8), parameter :: bspct0 = 0.03d0   ! initial range for bisection (the previous error is used after time 2) 0.03 for wage

    ! for convergence criteria (both in the steady state and the inner loop)
    integer, parameter :: logdiff = 0       ! =1 percetage difference
    ! for the initial guess of the value function in the inner loop
    integer, parameter :: vmatini = 1       ! =1 using the one in the previous iteration (used in innerloop)
    ! for the initial guess of the forecasting rules in the outer loop
    character(40) :: ckappakp = "./kappakp.txt"
    character(40) :: ckappaw  = "./kappaw.txt"
    character(40) :: ckappar  = "./kappar.txt"
    character(40) :: ckappal  = "./kappal.txt"
    ! forecasting rules chosen to be estimated
    integer, parameter :: fcsteqn = 1
    ! =1: kappakp and kappaw (kappar is obtained by the firm's FOC)
    ! =2: kappakp and kappar (kappaw is obtained by the firm's FOC)
    ! =3: kappakp, kappaw, and kappar

    ! for concavity degree adjustment (for grid of a and k)
    integer, parameter :: congrid = 1        ! grid option       ! =1 custermized
    real(8), parameter :: deggrida = 6.0d0   ! concavity degree adjustment
    real(8), parameter :: deggridk = 6.0d0   ! concavity degree adjustment
    ! Approximation of individual productivity
    integer, parameter :: nx = 10
    integer, parameter :: tauflag = 0 ! =1 Tauchen approximation, =0 Rouwenhorst instead
    real(8), parameter :: mx = 3.0d0
    ! Approximation of aggregate productivity
    integer, parameter :: nz = 5
    integer, parameter :: tauflagagg = 0
    real(8), parameter :: mz = 2.0d0
    integer, parameter :: nghz = 5

    !******* the number of grid points and the length of simulation *******!
    integer, parameter :: nb = 1
    integer, parameter :: ne = nb*nx
    integer, parameter :: nm = 40             ! NOTE: When linear interpolation is used, better to use a large number. With linear interpolation, this only slows down inner loop (which is not costly)
                                               ! It turns out that this is quite important for stability of outerloop convergence.
    real(8), parameter :: mpct = 0.15d0        ! percent range of aggregate capital (max dev of K in equilibrium is less than 10%)

    ! for simulation
    ! steady state
    integer, parameter :: NS = 1000000 ! number of samples
    integer, parameter :: T = 4 ! number of periods (quarters)
	integer, parameter :: yT = 1 ! number of periods (years)
    ! business cycles
    integer, parameter :: simT = 10000
    integer, parameter :: drop = 100
    integer, parameter :: simTT = simT+drop
    ! for impulse reponse analysis
    real(8), parameter :: shocksize = -0.02d0
    integer, parameter :: shockgrid = 1 ! =1 for negative shocks, =nz for positive shocks
    integer, parameter :: irfT = 100
    integer, parameter :: irfdrop = 50
    integer, parameter :: irfTT = irfT+irfdrop
    integer, parameter :: irfN = 4000

    !*******  structural parameters *******!
    real(8), parameter :: ALPHA = 0.36d0        ! Capital share
    real(8), parameter :: DELTA = 0.025d0       ! Depreciation rate
    real(8), parameter :: HBAR = 1.0d0/3.0d0    ! hours if working
    real(8), parameter :: RHOZ = 0.95d0         ! Persistence of aggregate shocks
    real(8), parameter :: SDINOVZ = 0.007d0     ! SD of innovations to aggregate shocks
    real(8), parameter :: RHOX = 0.9847d0        ! Persistence of idiosyncratic shocks (age 18-70)

    ! Not used for now
    real(8), parameter :: kappa0 = 0.0d0
    real(8), parameter :: phi = 0.0d0

    real(8), parameter :: betadev = 0.0d0
    real(8), parameter :: T1amin = 0.0d0       ! multiple of asset threshold for T1
    real(8), parameter :: blimit = 0.0d0

    ! for calibration by opt.f90
  	integer, parameter :: specflag = 1	! =1 HSV_HAT
  										! =2 HSV_HAN
  										! =3 LIN_HAT
  										! =4 LIN_HAN
  	integer, parameter :: npar = 3
  	integer, parameter :: hsvflag = 1 	! =1 HSV taxation, =0 linear taxation
    ! NOTE 190908: if we do calibration, just comment out the lines below

    ! Guner et al. 2014 with EITC
    real(8), parameter :: kappa1 = 0.053d0    ! Degree of progressivity
    real(8), parameter :: kappa2 = 1.0d0-(1.0d0-0.911d0)    ! Inverse of scale

    real(8), parameter :: tauss = 0.0d0 


    ! Calibrated using SMM
    ! HA-T
    real(8), parameter :: B0 = 0.6917d0
    real(8), parameter :: BETA = 0.9849d0

    real(8), parameter :: SDINOVX = 0.1262d0     ! SD of innovations to idiosyncratic shocks
    real(8), parameter :: Trscale = 0.1174d0
    real(8), parameter :: Transprog1 = 3.62d0
    real(8), parameter :: Transprog2 = 0.0d0
    real(8), parameter :: T0 = 0.0337d0
    !
    !! Note: Following is needed for HAN HAF models
    real(8), parameter :: T0Baseline = 0.0337d0    ! Manually insert the baseline value of T0




end module mod_parameters
