% impulse response functions
clear all;

load pfmat.mat;

RHOZ = 0.95;
SDINOVZ = 0.007;
shocksize = -0.02;
simt = 200;
drop = 50;
bigt = simt+drop; % length of simulated series

% fit splines
rk = nk-2;
cfmat = zeros(4,rk+1,nz);

for iz = 1:nz
    
    fcond = zeros(nk,1);

    for jz = 1:nz

        fcond = fcond + pzij(iz,jz)*fmat0(:,jz);

    end

    cfmat(:,:,iz) = spfit(invTk,fcond,rk,kgrid);
    
end

% do simulation with linear interpolation on z
% note that z_t is saved in simz(time), whereas c_t is saved in
% simc(time+1)
simz = zeros(bigt,1); % z_{t-1}: TFP in the previous period
simk = zeros(bigt,1); % k_{t-1}: capital at the end of the previous period, or at the beginning of the current period
simc = zeros(bigt+1,1); % c_t:
simy = zeros(bigt+1,1);
simh = zeros(bigt+1,1);
simx = zeros(bigt+1,1);
simw = zeros(bigt+1,1);
simr = zeros(bigt+1,1);

% determine the path of TFP shock
zgrid = exp(zgrid);
simz(1:drop+1) = 1.0; % at the steady state in the period 0
simz(drop+2) = max(exp(RHOZ*log(simz(drop+1))+shocksize),zgrid(1)); % shock hits in the period 1

for time=drop+2:bigt
    simz(time+1) = exp(RHOZ*log(simz(time)));
end

simk(1) = Kss;

crit = 1e-4;
x0 = [Hss Kss]';

kpint = zeros(2,1);
cint = zeros(2,1);
yint = zeros(2,1);
hint = zeros(2,1);
wint = zeros(2,1);
rint = zeros(2,1);

for time = 1:bigt
    
    iiz = gridlookup2(nz-2,zgrid,simz(time));

    if (iiz<nz)
        wz = (zgrid(iiz+1)-simz(time))/(zgrid(iiz+1)-zgrid(iiz));
    else
        iiz = iiz-1;
        wz = 0.0;
    end
    
    for iz = iiz:iiz+1
        
        znow = zgrid(iz);
        know = simk(time);

        [x0 rc] = csolve(@policy,x0,[],crit^2,100,cfmat(:,:,iz),kgrid,know,znow,ALPHA,DELTA,BETA,ETA,taul,Tr);
        n0 = x0(1);
        kp0 = x0(2);

        ynow = znow*know^ALPHA*n0^(1-ALPHA);
        w0 = (1-ALPHA)*znow*know^ALPHA*n0^(-ALPHA);
        r0 = ALPHA*znow*know^(ALPHA-1)*n0^(1-ALPHA) - DELTA;
        yterm = (1-DELTA)*know + (1-taul*(1-ALPHA))*ynow + Tr;
        cnow = yterm - kp0;
        
        kpint(iz-iiz+1) = kp0;
        cint(iz-iiz+1) = cnow;
        yint(iz-iiz+1) = ynow;
        hint(iz-iiz+1) = n0;
        wint(iz-iiz+1) = w0;
        rint(iz-iiz+1) = r0;
        
    end
    
    simk(time+1) = wz*kpint(1) + (1-wz)*kpint(2);
    simc(time+1) = wz*cint(1) + (1-wz)*cint(2);
    simy(time+1) = wz*yint(1) + (1-wz)*yint(2);
    simh(time+1) = wz*hint(1) + (1-wz)*hint(2);
    simw(time+1) = wz*wint(1) + (1-wz)*wint(2);
    simr(time+1) = wz*rint(1) + (1-wz)*rint(2);
    simx(time+1) = (wz*kpint(1) + (1-wz)*kpint(2))-(1-DELTA)*know;
    
end

% save data
simz = simz(drop+1:bigt);
simcons = simc(drop+2:bigt+1);
simy = simy(drop+2:bigt+1);
simh = simh(drop+2:bigt+1);
siminv = simx(drop+2:bigt+1);
simw = simw(drop+2:bigt+1);
simr = simr(drop+2:bigt+1);
simhy = simy./simh;

figure;
subplot(221);
plot(100*log(simz(1:80)/simz(end)));
subplot(222);
plot(100*log(simy(1:80)/simy(end)));
subplot(223);
plot(100*log(simh(1:80)/simh(end)));
subplot(224);
plot(100*log(simcons(1:80)/simcons(end)));

save repirf.mat simz simy simcons simh simhy siminv simw simr;