% Representative agent model
% Solved by using the policy function iteration method

% 20210209 QE revision
clear all;

ALPHA = 0.36;
DELTA = 0.025;
HBAR = 1/3;
taul = 0.0711/0.64; %0.279;
%taul = 0.0;
RHOZ = 0.95;
SDINOVZ = 0.007;

% NOTE: As for the rep agent model, the average labor tax rate should be 18.74% to be consistent with HAT model under the current calibration. Transfer is still 10.2% of GDP.% calibration targets
% note that TYss = 0 and taul = 0 yields to Hansen (1985)
rss = 0.01;
Hss = 0.782; %0.750;
%TYss = 0.106;
TYss = 0.044; %0.0;

BETA = 1/(1+rss);
GYss = taul*(1-ALPHA)-TYss;
KYss = ALPHA/(rss+DELTA);
ETA = (1-taul)*(1-ALPHA)/(1-DELTA*KYss-GYss)/Hss;
KHss = KYss^(1/(1-ALPHA));
Kss = KHss*Hss;
Yss = Kss/KYss;
Tr = TYss*Yss;
Css = (1.0-GYss)*Yss-DELTA*Kss;

nz = 5; %7;
nk = 25;
rk = nk-2;

[zgrid pzij] = Rouwenhorst(RHOZ,SDINOVZ,nz);
kgrid = linspace(0.85*Kss,1.15*Kss,nk)';
Tk = spbas(rk,kgrid);
invTk = inv(Tk);

% initial guess
% log-linear model is used as an educated initial guess
[f p] = linearsol(ALPHA,DELTA,BETA,ETA,taul,RHOZ,KYss,GYss,Kss,Css,Hss,kgrid,zgrid);
pause

for iz = 1:nz
    
    znow = exp(zgrid(iz));
    
    for ik = 1:nk
        
        know = kgrid(ik);
        
        fmat0(ik,iz) = 1.0/BETA/(Css*exp(f(1,1)*log(know/Kss)+f(1,2)*log(znow)));
        gmat0(ik,iz) = Kss*exp(p(1,1)*log(know/Kss)+p(1,2)*log(znow));
        hmat0(ik,iz) = Hss*exp(f(3,1)*log(know/Kss)+f(3,2)*log(znow));
        
    end
    
end
    

crit = 1e-4;
diff = 1e+4;
iter = 0;
%f = figure;

while (diff>crit)

    for iz = 1:nz
        
        znow = exp(zgrid(iz));
        
        fcond = zeros(nk,1);
        
        for jz = 1:nz
            
            fcond = fcond + pzij(iz,jz)*fmat0(:,jz);
            
        end
        
        cfmat = spfit(invTk,fcond,rk,kgrid);
        
        for ik = 1:nk

            know = kgrid(ik);

            x0 = [hmat0(ik,iz) gmat0(ik,iz)]';
            [x0 rc] = csolve(@policy,x0,[],crit^2,100,cfmat,kgrid,know,znow,ALPHA,DELTA,BETA,ETA,taul,Tr);

            n0 = x0(1);
            kp0 = x0(2);

            ynow = znow*know^ALPHA*n0^(1-ALPHA);
            yterm = (1-DELTA)*know + (1-taul*(1-ALPHA))*ynow + Tr;
            cnow = yterm - kp0;
            RHS = (1/cnow)*(1+ALPHA*ynow/know-DELTA);
%            disp([ik iz]);
%            disp([rc x0' ynow cnow]);
%            pause

            fmat1(ik,iz) = RHS;
            gmat1(ik,iz) = kp0;
            hmat1(ik,iz) = n0;

        end
        
    end
    
    iter = iter + 1;
    difff = max(max(abs(fmat1-fmat0)));
    diffg = max(max(abs(gmat1-gmat0)));
    diffh = max(max(abs(hmat1-hmat0)));
    diff = max([difff diffg diffh]);
    disp([iter diff]);
    
    fmat0 = fmat1;
    gmat0 = gmat1;
    hmat0 = hmat1;
        
end

save pfmat.mat ALPHA DELTA BETA ETA taul Tr Kss Css Yss Hss nk nz kgrid invTk zgrid pzij fmat0 gmat0 hmat0;