% simulation
clear all;

load pfmat.mat;

simt = 10000;
drop = 500;
bigt = simt+drop; % length of simulated series

% generate random numbers
rng(1,'twister');
cumpz= cumsum(pzij')';
izvec = zeros(bigt,1);
izvec(1) = 1;
for time = 1:bigt-1
    cumpzvec = cumpz(izvec(time),:);
    izvec(time+1) = sum(rand-cumpzvec>=0);
    izvec(time+1) = min(izvec(time+1)+1,nz);
end

% fit splines
rk = nk-2;
cfmat = zeros(4,rk+1,nz);

for iz = 1:nz
    
    fcond = zeros(nk,1);

    for jz = 1:nz

        fcond = fcond + pzij(iz,jz)*fmat0(:,jz);

    end

    cfmat(:,:,iz) = spfit(invTk,fcond,rk,kgrid);
    
end

% do simulation
simk = zeros(bigt,1); % k_{t-1}: capital at the end of the previous period, or at the beginning of the current period
simc = zeros(bigt,1); % c_t
simy = zeros(bigt,1);
simh = zeros(bigt,1);
simx = zeros(bigt,1);

simk(1) = Kss;
crit = 1e-4;

x0 = [Hss Kss]';

for time = 1:bigt-1
    
    znow = exp(zgrid(izvec(time)));
    know = simk(time);
    
    [x0 rc] = csolve(@policy,x0,[],crit^2,100,cfmat(:,:,izvec(time)),kgrid,know,znow,ALPHA,DELTA,BETA,ETA,taul,Tr);
    n0 = x0(1);
    kp0 = x0(2);

    ynow = znow*know^ALPHA*n0^(1-ALPHA);
    yterm = (1-DELTA)*know + (1-taul*(1-ALPHA))*ynow + Tr;
    cnow = yterm - kp0;
    
    simk(time+1) = kp0;
    simc(time+1) = cnow;
    simy(time+1) = ynow;
    simh(time+1) = n0;
    simx(time+1) = kp0-(1-DELTA)*know;
    
end

% business cycle statistics
Xss = DELTA*Kss;
ly = log(simy(drop+1:bigt));
lc = log(simc(drop+1:bigt));
lx = log(simx(drop+1:bigt));
lh = log(simh(drop+1:bigt));
lyh = log(simy(drop+1:bigt)./simh(drop+1:bigt));

T = length(ly);

xx = [ly lc lx lh lyh];
xf = hpfilter(xx,1600);
xf = xf(9:T-8,:);
temp = corr(xf);

stdvec = [std(xf(:,1)) std(xf(:,2:5))];
stdrelvec = [std(xf(:,1))*100 std(xf(:,2:5))./std(xf(:,1))];
corrvec = [temp(1,2:5) temp(4,5)]; % corr between y/h and h

save RA_simul stdrelvec corrvec;

% autocorrelation
rhoy = corr(xf(2:end,1),xf(1:end-1,1));
rhoc = corr(xf(2:end,2),xf(1:end-1,2));
rhox = corr(xf(2:end,3),xf(1:end-1,3));
rhoh = corr(xf(2:end,4),xf(1:end-1,4));
rhoyh = corr(xf(2:end,5),xf(1:end-1,5));

disp(' ');
disp(' standard deviations');
%disp('    sigy      sigc      sigx      sigh      sig(y/h)   sig(tr/y)');
disp('    sigy      sigc      sigx      sigh      sig(y/h)');
disp(stdvec*100);

disp(' ');
disp(' (relative) standard deviations');
disp('    sigy      sigc/sigy sigx/sigy sigh/sigy sig(y/h)/sigy');
disp(stdrelvec);

disp(' corr y with');
disp('    c         x         h        (y/h)');
disp([temp(1,2:5)]);
disp(sprintf(' corr h with (y/h): %6.5f',temp(4,5)));

disp(' ');
disp(' autocorrelation');
disp('    y         c         x         h         y/h');
disp([rhoy rhoc rhox rhoh rhoyh]);

ETA
BETA
Tr