! Sub program to compute reserve raises in Mui and Schoefer for "Implications of Government Transfers for Aggregate Labor Market Fluctuations"
! by Youngsoo Jang, Takeki Sunakawa and Minchul Yum

program prog_reserve_raises


	use mod_parameters
    use mod_utils
	use mod_calcss

    use mod_policies
    use sort
    
	implicit none

    ! increment of reservation raise base
    real(8), parameter :: increment = 0.005d0
    real(8), parameter :: max_res = 10.0d0
                                                
    integer, parameter :: consinterp = 1    ! 1: interpolate consumption; 0: use consumption resolved on kgrid
    
    real(8), parameter :: bisecttol = 1d-6
    
    real(8) bisectdist, valfix, valalt, netearn, apnow, EVtemp, vit, Nval, Wval, khigh, valtemp, xl, xr, x0, fl, fr, f0
    integer zeroind
    
    real(8) agrid(na), mgrid(nm), xgrid(nx), zgrid(nz), pxij(nx,nx), pzij(nz,nz), xss(nx), ALOW, atilde
    integer error
	integer ia, ja, ik, jk, im, jm, ix, jx, iz, jz, iterout, iteroutvec(100), i, foreind, izvec(simTT), time, iv
    real(8) tfp, w, r, Lss, Kss, Yss, TY, TrY, Trss, kgrid(nk), mumat(nk,nx), muk(nk), mustationary(nk,nx), ndist(nk,nx), Ks, Ls, EPratio, wss, rss
    real(8) TrInc5(5), TrW5(5), EmpW5(5), stdlogwage, stdlogearn, rhoxest, Tr12, Tr1, Tr2, Tr45
    real(8) vep1, vep2, vep3, vep4

    real(8) aggsim(simTT,10), disaggsim(simTT,nx,6)
    real(8), allocatable :: kappakp(:,:), kappaw(:,:), kappar(:,:), kappal(:,:), kappakpnew(:,:), kappawnew(:,:), kapparnew(:,:), kappalnew(:,:), rsquared(:,:)
    real(8) diffkp, diffw, diffr, diffl, diffout, damp, diffkpvec(100), diffwvec(100), diffrvec(100), difflvec(100), bspct, Vvec(na*nm*nx*nz)
    real(8) xghz(nghz), wghz(nghz), zvec(simTT)
    
    real(8) :: wmatrev(nk,nx), nmatrev(nk,nx), vmatssrev(na,nx), rvmat(nk,nx), evmatssrev(na,nx), cmatw(nk,nx), cmatn(nk,nx), apmatw(nk,nx), apmatn(nk,nx), cwmat(na,nx), cnmat(na,nx), Vmatss(na,nx) ! to compute the reserve raises
    real(8) :: rvwagemat(nk,nx)
    real(8) :: evrev(na), eval
    real(8) :: res_l, res_0, res_r, fval_l, fval_0, fval_r
    real(8) :: xnow, anow, Trw, taxw, kwhigh, awp, cnow, Trn, cwnow, cnnow

    real(8) :: temprvvec(nk*nx), rvvec(nk*nx), rvwagevec(nk*nx), EV(na)
    real(8), allocatable :: revpdf(:),revcdf(:),revdomain(:), revrs_x(:), trevpdf(:), arcelast(:)
    real(8), allocatable :: revwagepdf(:), revwagecdf(:), revwagedomain(:), revwage_x(:), trevwagepdf(:)
    logical :: vecmask(nk,nx)=.true.
    
    
    integer, parameter :: ra = na-2
    integer, parameter :: rm = nm-2
    real(8) invTa(ra,ra), invTm(rm,rm), cfa(4,ra+1), WORK(ra)    
    Integer IPIV(ra), INFO, iter, rvindex(nk*nx), kindex(nk*nx), eindex(nk*nx)
    integer i1, i2, j1, j2, k1, k2, l1, l2, pospts, revnum
    
    integer cto1, cto2, cro
    real(8) eptimeo
	! vsl
	integer, parameter :: method = VSL_RNG_METHOD_UNIFORM_STD
    integer errcode, brng, seed, maxsimind
    type(vsl_stream_state) :: stream
	real(8), allocatable :: murand(:), xrand(:)


    
    call system_clock(cto1, cro)

    
	maxsimind = nint(0.5d0*NS) 
	! for values of brng parameter, see https://software.intel.com/en-us/mkl-developer-reference-fortran-brng-parameter-definition#TBL10-2
	brng = vsl_brng_sfmt19937 ! use Mersenne Twister
	seed = 1234

	errcode = vslnewstream(stream,brng,seed)
	allocate(murand(NS),xrand(nx*maxsimind))
	errcode = vdrnguniform(method,stream,NS,murand,0.0d0,1.0d0)
	errcode = vdrnguniform(method,stream,nx*maxsimind,xrand,0.0d0,1.0d0)
    
    !print*, 'here1'
    call calcss(murand,xrand,agrid,kgrid,xgrid,pxij,xss,ALOW, &
		w,r,Lss,Kss,Yss,TY,TrY,Trss,mumat,ndist,Ks,Ls,EPratio,TrInc5,TrW5,EmpW5,stdlogwage,stdlogearn,rhoxest,error, &
		ALPHA,DELTA,HBAR,phi,B0,RHOX,SDINOVX,kappa1,kappa2,Trscale,Transprog1,Transprog2,T0, &
		tol,tolss,toldist,maxiter,dampss0,congrid,deggrida,deggridk,1,diagnum,logdiff,tauflag,calcma,linflag,howflag,hsvflag,mx, &
        cmatw,cmatn,cwmat,cnmat,Vmatss)
        

    ! store the staionary distribution
    mustationary = mumat

    
    
    print*,'Compute reservation raises'!! compute reservation raises for each state
    
        do ix = 1,nx
            xnow = xgrid(ix)
            
            ! compute conditional expected value functions on grid points
	        EV = 0.0d0
	        do jx = 1,nx
	            EV = EV + pxij(ix,jx)*reshape(Vmatss(:,jx),(/na/))
	        end do

	        do ik = 1,nk
                anow = kgrid(ik)
                
                call transfers(xnow,anow,r,w,0.0d0,Yss,Trscale,Transprog1,Transprog2,T0,Trn)
                atilde = (ALOW-Trn)/(1.0d0+r)
                
                if (consinterp==1) then
                    cwnow = lini(agrid,cwmat(:,ix),anow)
                    cnnow = lini(agrid,cnmat(:,ix),anow)
                else
                    cwnow = cmatw(ik,ix)                    
                    cnnow = cmatn(ik,ix)        
                end if
                    
                if (anow <= atilde) then
                    ! not possible to not work for any raise
                    res_0 = 0.0d0
                
                else
                    
	                ! fit spline
				    if (linflag==0) cfa = spfit(invTa,EV,ra,agrid)
                    
				    res_0 = bisectminwage(agrid,cfa,EV,xnow,anow,r,w,Yss,HBAR,Trscale,Transprog1,Transprog2,T0,B0,kappa1,kappa2,ALOW,tol,linflag,hsvflag)	
                    
                end if
                
                
                call transfers(xnow,anow,r,w,HBAR,Yss,Trscale,Transprog1,Transprog2,T0,Trw)
                call transfers(xnow,anow,r,w,0.0d0,Yss,Trscale,Transprog1,Transprog2,T0,Trn)
                call taxfunc(w*xnow*HBAR,anow,r,Yss,kappa1,kappa2,hsvflag,taxw)
                
                ! Converting reservation wage to reservation raise
                rvmat(ik,ix) = res_0 - 1.0d0
                
                if ((ndist(ik,ix)>0.0d0)) then
                    if (rvmat(ik,ix)>0.0d0) then
                        write(*, "('!!Work:    res_raise=', f7.3,'   a= ', F6.2, ' net e= ', F6.3, ' c_w= ', F6.3, ' c_n= ', F6.3, ' Trw= ', F6.3, ' Trn= ', F6.3)") rvmat(ik,ix), anow, w*xnow*HBAR - taxw, cwnow, cnnow, Trw, Trn                    
                        write(*, "('  Measure=', f10.8)") mumat(ik,ix)
                        
                    else
                        !write(*, "('  Work:    res_raise=', f7.3,'   a= ', F6.2, ' net e= ', F6.3, ' c_w= ', F6.3, ' c_n= ', F6.3, ' Trw= ', F6.3, ' Trn= ', F6.3)") rvmat(ik,ix), anow, w*xnow*HBAR - taxw, cwnow, cnnow, Trw, Trn                    
                    end if
                   
                    
                elseif ((ndist(ik,ix)==0.0d0)) then
                    if (rvmat(ik,ix)<0.0d0) then
                        write(*, "('!!No Work: res_raise=', f7.3,'   a= ', F6.2, ' net e= ', F6.3, ' c_w= ', F6.3, ' c_n= ', F6.3, ' Trw= ', F6.3, ' Trn= ', F6.3)") rvmat(ik,ix),anow, w*xnow*HBAR - taxw, cwnow, cnnow, Trw, Trn
                        write(*, "('  Measure=', f10.8)") mumat(ik,ix)
                        
                    else
                        !write(*, "('  No Work: res_raise=', f7.3,'   a= ', F6.2, ' net e= ', F6.3, ' c_w= ', F6.3, ' c_n= ', F6.3, ' Trw= ', F6.3, ' Trn= ', F6.3)") rvmat(ik,ix),anow, w*xnow*HBAR - taxw, cwnow, cnnow, Trw, Trn
                    end if
                    
                end if
                
            end do ! ix loop
            
        end do    
        
   

    print*, 'sort reservation raises' 
    
    temprvvec=pack(rvmat,vecmask)
    call bsort(nk*nx,temprvvec,rvvec,rvindex)
    
    Do i1=1,nk*nx
        j1 = mod(rvindex(i1),nk)
        If (j1 .eq. 0) then
            kindex(i1) = nk
        else
            kindex(i1) = j1
        end if
        
        IF (mod(rvindex(i1),nk*nx).eq.0) then
            k1 = (rvindex(i1)/(nk*nx))-1
        else
            k1 = rvindex(i1)/(nk*nx)
        end if
        j1 = rvindex(i1)-(nk*nx*k1)
        l1 = j1/nk
        If (mod(j1,nk) .eq. 0) then
            eindex(i1) = l1
        else
            eindex(i1) = l1+1
        end if
    end do
    
    print*, ' compute the CDF of reservation raises '
    pospts=count(mustationary.gt.0.0d0)    
    allocate(revpdf(pospts),revdomain(pospts))
    ! first generate the pdf of reservation raises
    k1=0
    Do i1=1,nk*nx
        ik=kindex(i1)
        ix=eindex(i1)
        If( mustationary(ik,ix).gt.0.0d0) then
            k1= k1 + 1
            revpdf(k1) = mustationary(ik,ix)
            revdomain(k1) = rvvec(i1)
        end if
    end do
    
    
    revnum= 1+int(3.0d0/increment)
    
    allocate(trevpdf(revnum),revcdf(revnum),revrs_x(revnum),arcelast(revnum))
    
    ! generate a grid for x-axis with an increment of "increment"
    revrs_x(1) = -1.0d0
    Do i1=2,revnum
        revrs_x(i1) = revrs_x(1)+dble(i1-1)*increment
    end do
   
    ! finally, generate the cdf of reservation raises
    trevpdf=0.0d0
    Do i1=1,pospts
        vep1= revdomain(i1)
        If (vep1 .le. revrs_x(1)) then
            trevpdf(1) = trevpdf(1)+revpdf(i1)
        else if (vep1 .le. revrs_x(revnum)) then
            Do j1=2,revnum
                If ((vep1>revrs_x(j1-1)).and.(vep1<=revrs_x(j1))) then
                    trevpdf(j1)= trevpdf(j1) + revpdf(i1)  
                end if
                if (revrs_x(j1)==0.0d0) then
                    zeroind = j1
                    
                end if                
            end do
        end if
    end do
    
    Do i1=1,revnum
        revcdf(i1)=sum(trevpdf(1:i1))
    end do
    
    
    do i1=1,revnum
        arcelast(i1) = ((revcdf(i1)-revcdf(zeroind))/revcdf(zeroind))/revrs_x(i1)
    end do
     
    open (unit=20,file="revcdf.csv",action="write",status="replace")        
    
        
        write (20,"(20000(f0.16,',',:))") revcdf    
    
    close(20)
    
    
    temprvvec=pack(rvwagemat,vecmask)
    call bsort(nk*nx,temprvvec,rvwagevec,rvindex)
    
    
    
    Do i1=1,nk*nx
        j1 = mod(rvindex(i1),nk)
        If (j1 .eq. 0) then
            kindex(i1) = nk
        else
            kindex(i1) = j1
        end if
        
        IF (mod(rvindex(i1),nk*nx).eq.0) then
            k1 = (rvindex(i1)/(nk*nx))-1
        else
            k1 = rvindex(i1)/(nk*nx)
        end if
        j1 = rvindex(i1)-(nk*nx*k1)
        l1 = j1/nk
        If (mod(j1,nk) .eq. 0) then
            eindex(i1) = l1
        else
            eindex(i1) = l1+1
        end if
    end do


    print*, ' compute the CDF of reservation wages '
    pospts=count(mustationary.gt.0.0d0)    
    allocate(revwagepdf(pospts),revwagedomain(pospts))
    ! first generate the pdf of reservation raises
    k1=0
    Do i1=1,nk*nx
        ik=kindex(i1)
        ix=eindex(i1)
        If( mustationary(ik,ix).gt.0.0d0) then
            k1= k1 + 1
            revwagepdf(k1) = mustationary(ik,ix)
            revwagedomain(k1) = rvwagevec(i1)
        end if
    end do
    
    
    revnum= 1+int(revwagedomain(pospts)/increment)
    allocate(trevwagepdf(revnum),revwagecdf(revnum),revwage_x(revnum))
    
    ! generate a grid for x-axis with an increment of "increment"
    revwage_x(1) = 0.0d0
    Do i1=2,revnum
        revwage_x(i1) = revwage_x(1)+dble(i1-1)*increment
    end do
   
    ! finally, generate the cdf of reservation wage
    trevwagepdf=0.0d0
    Do i1=1,pospts
        vep1= revwagedomain(i1)
        If (vep1 .le. revwage_x(1)) then
            trevwagepdf(1) = trevwagepdf(1)+revwagepdf(i1)
        else if (vep1 .le. revwage_x(revnum)) then
            Do j1=2,revnum
                If ((vep1>revwage_x(j1-1)).and.(vep1<=revwage_x(j1))) then
                    trevwagepdf(j1)= trevwagepdf(j1) + revwagepdf(i1)  
                end if
                !if (revrs_x(j1)==0.0d0) then
                !    zeroind = j1
                !    
                !end if                
            end do
        end if
    end do

    Do i1=1,revnum
        revwagecdf(i1)=sum(trevwagepdf(1:i1))
    end do



    call system_clock(cto2, cro)
    eptimeo = (dble(cto2-cto1)/dble(cro))/60.0d0
    write(*,"('  Total Elasped time (minutes) = ', F10.2)") eptimeo
    
    contains
    
    function netearnfn(resraise,xnow,anow,r,w,Y) result(netearn)

        real(8), intent(in) :: resraise, xnow, anow, r, w, Y
        real(8) netearn, taxw, Trw, Trn
                
        call transfers(xnow,anow,r,resraise*w,HBAR,Y,Trscale,Transprog1,Transprog2,T0,Trw)
        call transfers(xnow,anow,r,resraise*w,0.0d0,Y,Trscale,Transprog1,Transprog2,T0,Trn)
        call taxfunc(resraise*w*xnow*HBAR,anow,r,Y,kappa1,kappa2,hsvflag,taxw)
                    
        ! taxn = 0                    
        netearn = resraise*w*xnow*HBAR - taxw + Trw - Trn

    end function netearnfn
    
    
    function revwage(revw,xnow,anow,r,w,Y) result(netearn)

        real(8), intent(in) :: revw, xnow, anow, r, w, Y
        real(8) netearn, taxw, Trw, Trn
                
        call transfers(xnow,anow,r,revw,HBAR,Y,Trscale,Transprog1,Transprog2,T0,Trw)
        call transfers(xnow,anow,r,revw,0.0d0,Y,Trscale,Transprog1,Transprog2,T0,Trn)
        call taxfunc(revw*xnow*HBAR,anow,r,Y,kappa1,kappa2,hsvflag,taxw)
                    
        ! taxn = 0                    
        netearn = revw*xnow*HBAR - taxw + Trw - Trn

    end function revwage
    
    
             
    function bisectminwage(agrid,cfa,EV,xnow,anow,r,w,Yagg,HBAR,Trscale,Transprog1,Transprog2,T0,B0,kappa1,kappa2,ALOW,tol,linflag,hsvflag) result(raise0)

	    real(8), intent(in) :: agrid(:), cfa(:,:), EV(:), xnow, anow, r, w, Yagg, HBAR, Trscale, Transprog1,Transprog2, T0, B0, kappa1, kappa2, ALOW, tol
	    integer, intent(in) :: linflag, hsvflag
	    real(8) raise0, raiselow, raisehigh, WNdiffL, WNdiffH, WNdiff0, awp, anp, TWL, TNL, TWH, TNH, TW, TN, distance
        Real(8) :: cw, cn, minwage0
	    integer iteration
        
        call calcwnval(xnow,anow,r,w,Yagg,HBAR,Trscale,Transprog1,Transprog2,T0,B0,kappa1,kappa2,ALOW,agrid,cfa,EV,tol,linflag,0,hsvflag,awp,anp,TW,TN,cw,cn)
        
        if (TW > TN) then
            raiselow = 0.0d0    ! minimum for bisection   
            raisehigh = 1.0d0     ! maximum for bisection: upper bound of resv raise
        else
            raiselow = 1.0d0    ! minimum for bisection   
            raisehigh = max_res     ! maximum for bisection: upper bound of resv raise
        end if
        
        call calcwnval(xnow,anow,r,raiselow*w,Yagg,HBAR,Trscale,Transprog1,Transprog2,T0,B0,kappa1,kappa2,ALOW,agrid,cfa,EV,tol,linflag,0,hsvflag,awp,anp,TWL,TNL,cw,cn)
        WNdiffL = TWL - TNL
        
        call calcwnval(xnow,anow,r,raisehigh*w,Yagg,HBAR,Trscale,Transprog1,Transprog2,T0,B0,kappa1,kappa2,ALOW,agrid,cfa,EV,tol,linflag,0,hsvflag,awp,anp,TWH,TNH,cw,cn)
        WNdiffH = TWH - TNH
    
	    if (WNdiffL > 0.0d0) then

	        raise0 = raiselow
            !write(*,"('  TWL ', f8.4, '   TNL ', f8.4, '   WNdiffL ', f8.4)") TWL, TNL, WNdiffL

	    elseif (WNdiffH < 0.0d0) then

	        raise0 = raisehigh
            !print*, ' binding at the top' 

        else
            !print*, ' bisection' 
	    
	        if (WNdiffL*WNdiffH > 0.0d0) then

		        print *, "  Bisection prerequisite violated!"
	            write(*,"('  ik ', I4, '   ix ', I4)") ik, ix

		    else

			    ! bisection
			    distance = 1d+4
	            iteration = 0

			    do while (distance > tol*100.0d0)

	                raise0 = (raiselow+raisehigh)/2.0d0

				    call calcwnval(xnow,anow,r,raise0*w,Yagg,HBAR,Trscale,Transprog1,Transprog2,T0,B0,kappa1,kappa2,ALOW,agrid,cfa,EV,tol,linflag,0,hsvflag,awp,anp,TW,TN,cw,cn)

	                WNdiff0 = TW-TN

	                ! f(x0)*f(xl)
	                if (WNdiff0*WNdiffL > 0.0d0) then
	                    raiselow = raise0
	                else
	                    raisehigh = raise0
	                end if

	                distance = abs(raisehigh-raiselow)
	                iteration = iteration + 1
	                
                end do ! while loop for bisection
                
	        end if

        end if
        

    end function bisectminwage



    
    


end program prog_reserve_raises
