% calc business cycle statistics
% 20190715: z is continuous in the forecasting rule

clear all;
DELTA = 0.025;

% for HSV_HAT models
dir = './HSV_HAT/';
yvec = importdata(strcat(dir,'yvec.txt'));
cvec = importdata(strcat(dir,'cvec.txt'));
xvec = importdata(strcat(dir,'xvec.txt'));
lvec = importdata(strcat(dir,'lvec.txt'));
hvec = importdata(strcat(dir,'hvec.txt'));
zvec = importdata(strcat(dir,'zvec.txt'));
wvec = importdata(strcat(dir,'wvec.txt'));
rvec = importdata(strcat(dir,'rvec.txt'));

T = length(yvec);
yhvec = yvec./hvec;
rvec = rvec + DELTA;

so = log([yvec cvec xvec lvec hvec yhvec wvec]);
sd = hpfilter(so,1600);
sd = sd(9:T-8,:);
T0 = T-16;

std0 = std(sd);
stdrel_HAT = round([std0(1)*100 std0(2:6)/std0(1)],2)';
corr0 = corr(sd);
corr_HAT = round([corr0(1,2:6),corr0(5,6)],2)';

% Den Haan statistics
kappakp = importdata(strcat(dir,'kappakp.txt'));
kappaw = importdata(strcat(dir,'kappaw.txt'));
izvec = importdata(strcat(dir,'izvec.txt'));
mvec = importdata(strcat(dir,'mvec.txt'));

simT = size(izvec,1);
mhatvec = zeros(simT,1);
whatvec = zeros(simT,1);
rhatvec = zeros(simT,1);
lhatvec = zeros(simT,1);
mhatvec(1) = mvec(1);

for time = 1:simT
    
    iz = izvec(time);
    
    mhatvec(time+1) = exp(kappakp(1) + kappakp(2)*log(mhatvec(time)) + kappakp(3)*log(zvec(time)));
    whatvec(time)   = exp(kappaw(1)  + kappaw(2)*log(mhatvec(time)) + kappaw(3)*log(zvec(time)));
end

mhatvec = mhatvec(1:end-1);

DHmax = zeros(1,2);
DHmean = zeros(1,2);

DHmax(1)  = 100.0*max(abs(log(mvec./mhatvec)));
DHmean(1) = 100.0*sum(abs(log(mvec./mhatvec)))/simT;
DHmax(2)  = 100.0*max(abs(log(wvec./whatvec)));
DHmean(2) = 100.0*sum(abs(log(wvec./whatvec)))/simT;

%tabulate(zvec);

mpvec = mvec(2:end);
mvec = mvec(1:end-1);
wvec = wvec(1:end-1);
zvec = zvec(1:end-1);

X = ones(T-1,3);
X(:,2) = log(zvec);
X(:,3) = log(mvec);

[b,~,~,~,stats] = regress(log(mpvec),X);
r2kp = stats(1);
temp1_HAT = [round(kappakp,4)',round(r2kp,5),round([DHmean(1),DHmax(1)],3)];

[b,~,~,~,stats] = regress(log(wvec),X);
r2w = stats(1);
temp2_HAT = [round(kappaw,4)',round(r2w,5),round([DHmean(2),DHmax(2)],3)];

% for HSV_HAN models
dir = './HSV_HAN/';
yvec = importdata(strcat(dir,'yvec.txt'));
cvec = importdata(strcat(dir,'cvec.txt'));
xvec = importdata(strcat(dir,'xvec.txt'));
lvec = importdata(strcat(dir,'lvec.txt'));
hvec = importdata(strcat(dir,'hvec.txt'));
zvec = importdata(strcat(dir,'zvec.txt'));
wvec = importdata(strcat(dir,'wvec.txt'));
rvec = importdata(strcat(dir,'rvec.txt'));

T = length(yvec);
yhvec = yvec./hvec;
rvec = rvec + DELTA;

so = log([yvec cvec xvec lvec hvec yhvec wvec]);
sd = hpfilter(so,1600);
sd = sd(9:T-8,:);
T0 = T-16;

std0 = std(sd);
stdrel_HAN = round([std0(1)*100 std0(2:6)/std0(1)],2)';
corr0 = corr(sd);
corr_HAN = round([corr0(1,2:6),corr0(5,6)],2)';

% Den Haan statistics
kappakp = importdata(strcat(dir,'kappakp.txt'));
kappaw = importdata(strcat(dir,'kappaw.txt'));
izvec = importdata(strcat(dir,'izvec.txt'));
mvec = importdata(strcat(dir,'mvec.txt'));

simT = size(izvec,1);
mhatvec = zeros(simT,1);
whatvec = zeros(simT,1);
rhatvec = zeros(simT,1);
lhatvec = zeros(simT,1);
mhatvec(1) = mvec(1);

for time = 1:simT
    
    iz = izvec(time);
    
    mhatvec(time+1) = exp(kappakp(1) + kappakp(2)*log(mhatvec(time)) + kappakp(3)*log(zvec(time)));
    whatvec(time)   = exp(kappaw(1)  + kappaw(2)*log(mhatvec(time)) + kappaw(3)*log(zvec(time)));
end

mhatvec = mhatvec(1:end-1);

DHmax = zeros(1,2);
DHmean = zeros(1,2);

DHmax(1)  = 100.0*max(abs(log(mvec./mhatvec)));
DHmean(1) = 100.0*sum(abs(log(mvec./mhatvec)))/simT;
DHmax(2)  = 100.0*max(abs(log(wvec./whatvec)));
DHmean(2) = 100.0*sum(abs(log(wvec./whatvec)))/simT;

%tabulate(zvec);

mpvec = mvec(2:end);
mvec = mvec(1:end-1);
wvec = wvec(1:end-1);
zvec = zvec(1:end-1);

X = ones(T-1,3);
X(:,2) = log(zvec);
X(:,3) = log(mvec);

[b,~,~,~,stats] = regress(log(mpvec),X);
r2kp = stats(1);
temp1_HAN = [round(kappakp,4)',round(r2kp,5),round([DHmean(1),DHmax(1)],3)];

[b,~,~,~,stats] = regress(log(wvec),X);
r2w = stats(1);
temp2_HAN = [round(kappaw,4)',round(r2w,5),round([DHmean(2),DHmax(2)],3)];

% for HSV_HAF models
dir = './HSV_HAF/';
yvec = importdata(strcat(dir,'yvec.txt'));
cvec = importdata(strcat(dir,'cvec.txt'));
xvec = importdata(strcat(dir,'xvec.txt'));
lvec = importdata(strcat(dir,'lvec.txt'));
hvec = importdata(strcat(dir,'hvec.txt'));
zvec = importdata(strcat(dir,'zvec.txt'));
wvec = importdata(strcat(dir,'wvec.txt'));
rvec = importdata(strcat(dir,'rvec.txt'));

T = length(yvec);
yhvec = yvec./hvec;
rvec = rvec + DELTA;

so = log([yvec cvec xvec lvec hvec yhvec wvec]);
sd = hpfilter(so,1600);
sd = sd(9:T-8,:);
T0 = T-16;

std0 = std(sd);
stdrel_HAF = round([std0(1)*100 std0(2:6)/std0(1)],2)';
corr0 = corr(sd);
corr_HAF = round([corr0(1,2:6),corr0(5,6)],2)';

% Den Haan statistics
kappakp = importdata(strcat(dir,'kappakp.txt'));
kappaw = importdata(strcat(dir,'kappaw.txt'));
izvec = importdata(strcat(dir,'izvec.txt'));
mvec = importdata(strcat(dir,'mvec.txt'));

simT = size(izvec,1);
mhatvec = zeros(simT,1);
whatvec = zeros(simT,1);
rhatvec = zeros(simT,1);
lhatvec = zeros(simT,1);
mhatvec(1) = mvec(1);

for time = 1:simT
    
    iz = izvec(time);
    
    mhatvec(time+1) = exp(kappakp(1) + kappakp(2)*log(mhatvec(time)) + kappakp(3)*log(zvec(time)));
    whatvec(time)   = exp(kappaw(1)  + kappaw(2)*log(mhatvec(time)) + kappaw(3)*log(zvec(time)));
end

mhatvec = mhatvec(1:end-1);

DHmax = zeros(1,2);
DHmean = zeros(1,2);

DHmax(1)  = 100.0*max(abs(log(mvec./mhatvec)));
DHmean(1) = 100.0*sum(abs(log(mvec./mhatvec)))/simT;
DHmax(2)  = 100.0*max(abs(log(wvec./whatvec)));
DHmean(2) = 100.0*sum(abs(log(wvec./whatvec)))/simT;

%tabulate(zvec);

mpvec = mvec(2:end);
mvec = mvec(1:end-1);
wvec = wvec(1:end-1);
zvec = zvec(1:end-1);

X = ones(T-1,3);
X(:,2) = log(zvec);
X(:,3) = log(mvec);

[b,~,~,~,stats] = regress(log(mpvec),X);
r2kp = stats(1);
temp1_HAF = [round(kappakp,4)',round(r2kp,5),round([DHmean(1),DHmax(1)],3)];

[b,~,~,~,stats] = regress(log(wvec),X);
r2w = stats(1);
temp2_HAF = [round(kappaw,4)',round(r2w,5),round([DHmean(2),DHmax(2)],3)];

load ./JSYrep/RA_simul stdrelvec corrvec;

disp(' ');
disp(' Table 4: Volatility of aggregate variables');
disp(' ');
disp([stdrel_HAT stdrel_HAN stdrel_HAF round([stdrelvec(1:3),NaN,stdrelvec(4:5)],2)']);

disp(' ');
disp(' Table 5: Cyclicality of aggregate variables');
disp(' ');
disp([corr_HAT corr_HAN corr_HAF round([corrvec(1:2),NaN,corrvec(3:5)],2)']);

disp(' ');
disp(' Table A2: Estimates and accuracy of forecasting rules');
disp(' ');
disp(sprintf('  %3.4f %1.4f %1.4f %1.5f %1.3f %1.3f',temp1_HAT));
disp(sprintf('  %3.4f %1.4f %1.4f %1.5f %1.3f %1.3f',temp2_HAT));
disp(sprintf('  %3.4f %1.4f %1.4f %1.5f %1.3f %1.3f',temp1_HAN));
disp(sprintf('  %3.4f %1.4f %1.4f %1.5f %1.3f %1.3f',temp2_HAN));
disp(sprintf('  %3.4f %1.4f %1.4f %1.5f %1.3f %1.3f',temp1_HAF));
disp(sprintf('  %3.4f %1.4f %1.4f %1.5f %1.3f %1.3f',temp2_HAF));